# -*- coding: utf-8 -*-
"""FCA library"""

from fcatng.concept import Concept
from fcatng.concept_system import ConceptSystem
from fcatng.concept_lattice import ConceptLattice
from fcatng.context import Context
from fcatng.mvcontext import ManyValuedContext
from fcatng.scale import Scale
from fcatng.implication import Implication

from fcatng.algorithms import (norris, compute_covering_relation,
                            scale_mvcontext, compute_dg_basis)
from fcatng.readwrite import (read_txt, read_cxt, write_cxt, write_dot,
                           read_mv_txt, read_xml, write_xml, write_mv_txt,
                           uread_cxt, uwrite_cxt)
from fcatng.algorithms.filters import (filter_concepts, compute_estability,
compute_istability, compute_separation_index, 
compute_probability, compute_index)



from .algorithms import norris

class ConceptLattice(object):
    """ConceptLattice class

    Examples
    ========
    
    >>> from fcatng import (Context, Concept)
    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = ['1', '2', '3', '4']
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> cl = ConceptLattice(c)
    >>> print cl
    ([], M)
    (['1'], ['a', 'd'])
    (['2'], ['a', 'c'])
    (['1', '2'], ['a'])
    (['3', '4'], ['b', 'c'])
    (['2', '3', '4'], ['c'])
    (G, [])
    (['4'], ['b', 'c', 'd'])
    (['1', '4'], ['d'])
    >>> print cl.parents(cl[5]) == set((cl[6],))
    True
    >>> print cl.children(cl[6]) == set((cl[5], cl[3], cl[8]))
    True

    """
    def __init__(self, context, builder=norris):
        (self._concepts, self._parents) = norris(context)
        self._bottom_concept = [c for c in self._concepts if not self.ideal(c)][0]
        self._top_concept = [c for c in self._concepts if not self.filter(c)][0]
        self._context = context
    
    def get_context(self):
        return self._context
    
    context = property(get_context)
    
    def get_top_concept(self):
        # TODO: change
        return self._top_concept
    top_concept = property(get_top_concept)

    def get_bottom_concept(self):
        # TODO: change
        return self._bottom_concept

    bottom_concept = property(get_bottom_concept)

    def filter(self, concept):
        # TODO: optimize
        return [c for c in self._concepts if concept.intent > c.intent]

    def ideal(self, concept):
        # TODO: optimize
        return [c for c in self._concepts if c.intent > concept.intent]

    def __len__(self):
        return len(self._concepts)

    def __getitem__(self, key):
        return self._concepts[key]

    def __contains__(self, value):
        return value in self._concepts

    def __str__(self):
        s = ""
        for c in self._concepts:
            s = s + "%s\n" % str(c)
        return s[:-1]

    def index(self, concept):
        return self._concepts.index(concept)
    
    def parents(self, concept):
        return self._parents[concept]

    def children(self, concept):
        return set([c for c in self._concepts if concept in self.parents(c)])
    
if __name__ == "__main__":
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
"""
Holds class for many valued context
"""

from fcatng import Context

class ManyValuedContext(Context):
    """
    Many valued context.

    Examples
    ========

    Create a context.

    >>> t = [[2, 1, 2, 3],\
              [1, 2, 3, 1],\
              [2, 1, 3, 1],\
              [2, 3, 3, 2]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = ManyValuedContext(t, objs, attrs)
    >>> c[0][0]
    2
    >>> c[0][2]
    2
    >>> 1 in c.objects
    True
    >>> 'f' in c.attributes
    False
    >>> for o in c:
    ...     print o
    ...     break
    ...
    [2, 1, 2, 3]

    Class emulates container type.

    >>> len(c)
    4
    >>> c[1]
    [1, 2, 3, 1]
    """

    def __init__(self, table=[], objects=[], attributes=[]):
        """Create a many valued context from table and list of objects, list
        of attributes
        
        table - list of attribute values lists
        objects - the list of objects
        attributes - the list of attributes 
        """
        if len(table) != len(objects):
            raise ValueError("Number of objects (=%i) and number of cross table"
                   " rows(=%i) must agree" % (len(objects), len(table)))
        elif (len(table) != 0) and len(table[0]) != len(attributes):
            raise ValueError("Number of attributes (=%i) and number of cross table"
                    " columns (=%i) must agree" % (len(attributes),
                        len(table[0])))

        self._table = table
        self._objects = objects
        self._attributes = attributes

    def get_objects(self):
        return self._objects

    objects = property(get_objects)

    def get_attributes(self):
        return self._attributes
        
    attributes = property(get_attributes)

    def extract_subcontext(self, attribute_names):
        """Create a subcontext with only indicated attributes"""
        return ManyValuedContext(self._extract_subtable(attribute_names),
                                self.objects,
                                attribute_names)
                                
    def extract_subcontext_by_condition(self, condition):
        """Extract a subcontext containing only objects that satisfy the
        condition.
        
        Keyword arguments:
        condition(object_index) -- a function that takes an an object index and
            returns a Boolean value
        
        """
        object_names, table = self._extract_subtable_by_condition(condition)
        return ManyValuedContext(table, object_names, self.attributes)
                                
    def extract_subcontext_by_attribute_values(self, values):
        """Extract a subcontext containing only objects with certain attribute
        values.
        
        Keyword arguments:
        values -- an attribute-value dictionary
        
        """
        object_names, table = self._extract_subtable_by_attribute_values(values)
        return ManyValuedContext(table, object_names, self.attributes)
                            

    ############################
    # Emulating container type #
    ############################

    def __len__(self):
        return len(self._table)

    def __getitem__(self, key):
        return self._table[key]

    ############################

if __name__ == "__main__":
    import doctest
    doctest.testmod()



""" Holds class for context with partially given exmaples (partial context) """

import fcatng.algorithms
from fcatng import Context
from fcatng.algorithms.closure_operators import aprime, oprime
from fcatng.algorithms.closure_operators import simple_closure as closure
from .compare_context import subseteq_table
from fcatng.algorithms.dg_basis import compute_partial_dg_basis
from copy import copy, deepcopy

class PartialContext(object):
    """
    A Partial Context consists of two sets: *objects* and *attributes*
    and of a ternary (x, q) relation between them.
    """
    def __init__(self, x_table, q_table, objects, attributes):
        """
        Create a partial context from cross and question tables
        and list of objects, list of attributes.

        cross_table_x - the list of bool lists representing first context
            (only x in context)
        cross_table_q - the list of bool lists representing second context
            (only questions(q) in context)
                cross_table_x \subseteq cross_table_q
        objects - the list of objects
        attributes - the list of attributes
        """
        if not subseteq_table(x_table, q_table):
            raise ValueError('x table should be \subseteq q table')
        else:
            self.objects = copy(objects)
            self.attributes = copy(attributes)
            self.x_context = Context(x_table, objects[:], self.attributes[:])
            self.q_context = Context(q_table, objects[:], self.attributes[:])

    def __deepcopy__(self, memo):
        return PartialContext(deepcopy(self.x_context._table, memo),
                              deepcopy(self.q_context._table, memo),
                              self.objects[:],
                              self.attributes[:])

    def transpose(self):
        """ Return new partial context with transposed x and q contexts """
        new_objects = self.attributes[:]
        new_attributes = self.objects[:]
        new_x_table = []
        new_q_table = []
        for j in range(len(self.attributes)):
            x_line = []
            q_line = []
            for i in range(len(self.objects)):
                x_line.append(self.x_context._table[i][j])
                q_line.append(self.q_context._table[i][j])
            new_x_table.append(x_line)
            new_q_table.append(q_line)
        return PartialContext(new_x_table, new_q_table, new_objects, new_attributes)

    def __repr__(self):
        output = ", ".join([str(a) for a in self.attributes]) + "\n"
        output += ", ".join([str(o) for o in self.objects]) + "\n"
        for o in self.objects:
            output += "\n"
            output += self._intent_to_str(self.x_context.get_object_intent(o),
                                          self.q_context.get_object_intent(o))
        return output
        
    def _intent_to_str(self, xintent, qintent):
        line = ""
        for a in self.attributes:
            if a in xintent:
                line += "X"
            elif a in qintent:
                line += "?"
            else:
                line += "."
        return line
        

    def xq_aclosure(self, aset):
        """ +? closure for Partial context """
        oset = aprime(aset, self.x_context)
        return oprime(oset, self.q_context)

    def get_attribute_implications(self,
                                   basis=compute_partial_dg_basis,
                                   confirmed=[],
                                   cond=lambda x: True):
        return basis(self, imp_basis=confirmed, cond=cond)

    def add_object_with_intent(self, intent, name):
        if not intent[0] <= intent[1]:
            raise ValueError('inconsistent partial example')
        self.x_context.add_object_with_intent(intent[0], name)
        self.q_context.add_object_with_intent(intent[1], name)
        self.objects.append(name)
        
    def complete(self, implications):
        for o in self.objects:
            xintent = self.x_context.get_object_intent(o)
            new_xintent = closure(xintent, implications)
            qintent = self.q_context.get_object_intent(o)
            if not new_xintent <= qintent:
                # TODO: undo the modifications
                raise ValueError(
                    'implications are inconsistent with the partial context')
            self.x_context.set_object_intent(new_xintent, o)
            new_qintent = set([a for a in qintent
                                 if a in new_xintent or
                                 closure(new_xintent | set([a]), implications)
                                                                    <= qintent
                             ])
            self.q_context.set_object_intent(new_qintent, o)
            # TODO: Remove printing
            if xintent != new_xintent or qintent != new_qintent:
                print('Object', o)
                print(self._intent_to_str(xintent, qintent))
                print('completed to ')
                print(self._intent_to_str(new_xintent, new_qintent))
            
    def intents(self):
        for obj in self.objects:
            yield (self.x_context.get_object_intent(obj),
                   self.q_context.get_object_intent(obj))
        


#!/usr/bin/env python
# encoding: utf-8
import unittest

from exploration import (AttributeExploration, ExplorationDB, 
                        NotCounterexample, IllegalContextModification)
import fcatng

from fcatng import Implication
from fcatng.algorithms.closure_operators import simple_closure as closure


class GenericExpert(object):
    
    def __init__(self, basis):
        self.basis = basis
        
    def is_valid(self, imp):
        return imp.conclusion <= closure(imp.premise, self.basis)
            
    def explore(self, exploration):
        while exploration.get_open_implications():
            imp = exploration.get_open_implications()[0]
            if self.is_valid(imp):
                exploration.confirm_implication(imp)
            else:
                exploration.reject_implication(imp)
        
    def provide_counterexample(self, imp):
        return (str(imp.premise), closure(imp.premise, self.basis))


class DummyPerfectExpert(GenericExpert):
        
    def __init__(self):
        GenericExpert.__init__(self, list())

        
class DummyFoolishExpert(object):

    def provide_counterexample(self, imp):
        return ('test', imp.premise | imp.conclusion)
        

class ExplorationTest(unittest.TestCase):
    def setUp(self):
        table = [[True, False, False, True],\
                 [True, False, True, False],\
                 [False, True, True, False],\
                 [False, True, True, True]]
        objs = ['1', '2', '3', '4']
        attrs = ['a', 'b', 'c', 'd']
        cxt = fcatng.Context(table, objs, attrs)
        self.db = ExplorationDB(cxt, list())
        
        
class InitializationTest(ExplorationTest):
    
    def test_with_perfect_expert(self):
        expert = DummyPerfectExpert()
        exploration = AttributeExploration(self.db, expert)
        self.assertEqual(len(self.db.open_implications), 3,
                         "Wrong number of attribute implications")
        
    def test_with_foolish_expert(self):
        expert = DummyFoolishExpert()
        exploration = AttributeExploration(self.db, expert)
        self.assertEqual(len(self.db.open_implications), 3,
                         "Wrong number of attribute implications")
                         
        
class ConfirmingImplicationsTest(ExplorationTest):
    
    def test_implication_confirm(self):
        expert = DummyPerfectExpert()
        exploration = AttributeExploration(self.db, expert)
        exploration.confirm_implication(self.db.open_implications[0])
        self.assertEqual(len(self.db.base), 1)
        self.assertEqual(len(self.db.open_implications), 2)
        
        
class RejectingImplicationTest(ExplorationTest):
    
    def test_correct_implication_reject(self):
        expert = DummyPerfectExpert()
        exploration = AttributeExploration(self.db, expert)
        exploration.reject_implication(self.db.open_implications[0])
        self.assertEqual(len(self.db._cxt.objects), 5)
        
    def test_incorrect_implication_reject(self):
        expert = DummyFoolishExpert()
        exploration = AttributeExploration(self.db, expert)
        self.assertRaises(
            NotCounterexample,
            exploration.reject_implication,
            self.db.open_implications[0]    
        )
        self.assertEqual(len(self.db._cxt.objects), 4)
        
        
class AddExampleTest(ExplorationTest):
    
    def test_add_correct_example(self):
        expert = DummyPerfectExpert()
        exploration = AttributeExploration(self.db, expert)
        exploration.confirm_implication(self.db.open_implications[0])
        imp = self.db.open_implications[0]
        intent = imp.premise | imp.conclusion
        self.db.add_example('test', intent)
        self.assertEqual(len(self.db._cxt.objects), 5)
        
    def test_add_incorrect_example(self):
        expert = DummyPerfectExpert()
        exploration = AttributeExploration(self.db, expert)
        exploration.confirm_implication(self.db.open_implications[0])
        imp = self.db.base[0]
        intent = imp.premise
        self.assertRaises(
            IllegalContextModification,
            self.db.add_example,
            'test', intent
        )
        self.assertEqual(len(self.db._cxt.objects), 4)
        

equilateral = 'equilateral'
isosceles = 'isosceles'
acute = 'acute-angled'
right = 'right-angled'


class TrianglesTest(unittest.TestCase):

    def setUp(self):
        table = [[False, True, False, False],
                 [False, False, True, False],
                 [False, False, False, True]]
        objs = [isosceles, acute, right]
        attrs = [equilateral, isosceles, acute, right]
        cxt = fcatng.Context(table, objs, attrs)
        self.db = ExplorationDB(cxt, list())
        
    def test_all(self):
        basis = [
                Implication(set([acute, right]), set([equilateral, isosceles])),
                Implication(set([equilateral]), set([isosceles, acute]))
                ]
        expert = GenericExpert(basis)
        exploration = AttributeExploration(self.db, expert)
        expert.explore(exploration)
        self.assertEqual(len(self.db._cxt.objects), 6)
        self.assertEqual(len(self.db.base), len(basis))
        
        
if __name__ == '__main__':
    unittest.main()



import fcatng

from fcatng.algorithms import closure_operators


def kclosure(s, k, cxt):
    """Return the closure of s in cxt restricted to the first k attributes."""
    closure = set(cxt.attributes[:k])
    for o in cxt.examples():
        if s == closure:
            break
        if s <= o:
            closure &= o
    return closure


# todo: optimize!
def compare_tuples(p, r):
    diff = p[1] ^ r[1]
    if diff:
        return 1 if sorted(diff)[0] in p[1] else -1
    else:
        return 0


def canonical_basis(cxt, close=closure_operators.simple_closure):
    preclosed = [(set(cxt.objects), set())]
    basis = []
    # each preclosed is (extent, intent) or (extent, premise, implication)
    # assuming that premise and implication.premise are the same object
    for i in range(len(cxt.attributes)):
        preclosed, basis = update_preclosed(i, cxt, preclosed, close)        
    return basis


def update_preclosed(i, cxt, preclosed,
                     close=closure_operators.simple_closure):

        m = cxt.attributes[i]
        extent = cxt.get_attribute_extent_by_index(i)
        context_closure = lambda s: kclosure(s, i + 1, cxt)
    
        old_stable_impl = []    # stores implications
        new_stable_impl = []    # stores implications
        min_mod_impl = []       # stores implications
        
        non_min_mod = []        # stores triples
        mod_extra = []          # stores pairs or triples

        new_preclosed = []

        for p in preclosed:
            if p[0] <= extent:  # p[1] -> m holds
                if is_concept(p):
                    process_modified_concept(p, m, min_mod_impl, mod_extra,
                                             new_preclosed)
                else:
                    process_modified_implication(p, m, min_mod_impl,
                                                 non_min_mod,
                                                 new_preclosed)
            else:               # p[1] -> m does not hold
                new_preclosed.append(p)
                if is_concept(p):       # p[1] remains closed
                    process_stable_concept(p, m, extent, new_stable_impl,
                                           new_preclosed,
                                           context_closure)
                else:                   # p[1] remains pseudo-closed
                    old_stable_impl.append(p[2])

        basis = old_stable_impl + new_stable_impl + min_mod_impl
        n = len(basis)
        basis += [p[2] for p in non_min_mod]
        for j in range(len(non_min_mod) - 1, -1, -1):
            impl = non_min_mod[j][2]
            del basis[n + j]
            premise = impl.premise
            premise |= close(premise, basis)    # sic! changing impl.premise
            if premise != impl.get_conclusion():
                basis.append(impl)
                mod_extra.append(non_min_mod[j])

        mod_extra.sort(cmp=compare_tuples)

        return new_preclosed + mod_extra, basis


def process_stable_concept(p, m, extent, new_stable_impl, new_preclosed,
                           closure):
    # p is of the form (extent, intent)
    new_extent = p[0] & extent
    new_premise = p[1].copy()
    new_premise.add(m)
    for i in new_stable_impl:
        if not i.is_respected(new_premise):
            break
    else:
        new_conclusion = closure(new_premise)
        if new_conclusion == new_premise:
            new_preclosed.append((new_extent, new_premise))
        else:
            impl = fcatng.Implication(new_premise, new_conclusion)
            new_stable_impl.append(impl)
            new_preclosed.append((new_extent, impl.premise, impl))


def process_modified_implication(p, m, min_mod_impl, non_min_mod,
                                 new_preclosed):
    # p is of the form (extent, premise, implication)
    p[2].get_conclusion().add(m)
    for i in min_mod_impl:
        if i.premise <= p[1]:   # p[1] is no longer preclosed
            p[1].add(m)         # assuming that p[1] is p[2].premise
            non_min_mod.append(p)
            break
    else:                       # p[1] remains psuedo-closed
        min_mod_impl.append(p[2])
        new_preclosed.append(p)


def process_modified_concept(p, m, min_mod_impl, mod_concepts, new_preclosed):
    # p is of the form (extent, intent)
    for i in min_mod_impl:
        if i.premise <= p[1]:   # p[1] is no longer preclosed
            break
    else:                       # p[1] becomes psuedo-closed
        impl = fcatng.Implication(p[1].copy(), p[1].copy())
        impl.get_conclusion().add(m)
        min_mod_impl.append(impl)
        new_preclosed.append((p[0], impl.premise, impl))
    p[1].add(m)
    mod_concepts.append(p)
    

def is_concept(p):
    return len(p) == 2


if __name__ == '__main__':
    import time
    import sys
    
    
    def timeit(algo, cxt):
        start = time.time()
        basis = algo(cxt)
        end = time.time()
        print((len(basis), end - start))
        return basis, end - start
        

    cxt = fcatng.read_cxt(sys.argv[1])
    print('Ganter:')
    timeit(fcatng.compute_dg_basis, cxt)
    print('Incremental:')
    timeit(canonical_basis, cxt)



from fcatng import ConceptSystem

# import separation
# import probability
# import stability

def compute_index(lattice, function, name):
    indexes = function(lattice)
    
    for concept in indexes.items():
        if concept[0].meta:
            concept[0].meta[name] = concept[1]
        else: 
            concept[0].meta = {name : concept[1]}

def filter_concepts(lattice, function, mode, opt=1):
    """Return new concept system, filtered by function according to the mode.
    
    Modes:
    --- "part" - part of initial concept lattice
    --- "abs" - absolute value of the concepts in resulting concept system
    --- "value" - value of the index
    
    Additionaly add attribute, containing inforamtion about indexes, to the new lattice
    """
    def _filter_value(lattice, indexes, value):
        filtered_concepts = [item for item in indexes.items() if item[1]>=value]
        return ConceptSystem([c[0] for c in filtered_concepts])
    
    def _filter_abs(lattice, indexes, n):
        cmp_ = lambda x,y: cmp(x[1], y[1])
        sorted_indexes = sorted(indexes.items(), cmp_, reverse=True)
        filtered_concepts = sorted_indexes[:int(n)]
            
        return ConceptSystem([c[0] for c in filtered_concepts])
    
    def _filter_part(lattice, indexes, part):
        n = int(len(lattice) * part)
        cmp_ = lambda x,y: cmp(x[1], y[1])
        sorted_indexes = sorted(indexes.items(), cmp_, reverse=True)
        filtered_concepts = sorted_indexes[:n]
        
        values = sorted_indexes
        eps = values[n-2][1]-values[n-1][1]
        
        other_concepts = sorted_indexes[n:]
        for concept in other_concepts:
            if abs(concept[1] - values[n][1]) < eps:
                filtered_concepts.append(concept)
            
        return ConceptSystem([c[0] for c in filtered_concepts])
    
    indexes = function(lattice)
    if indexes:
        if mode == "part":
            ret = _filter_part(lattice, indexes, opt)
        elif mode == "abs":
            ret = _filter_abs(lattice, indexes, opt)
        elif mode == "value":
            ret = _filter_value(lattice, indexes, opt)
    return ret

def compute_extent_size(lattice):
    number_of_objects = len(lattice.context.objects)
    extent_size_index = {}
    
    for c in lattice:
        extent_size_index[c] = len(c.extent) / number_of_objects
            
    return extent_size_index

def compute_intent_size(lattice):
    number_of_attributes = len(lattice.context.attributes)
    intent_size_index = {}
    
    for c in lattice:
        intent_size_index[c] = len(c.intent) / number_of_attributes
            
    return intent_size_index

from math import exp, log

def compute_probability(lattice):
    
    def get_intent_probability(B, p_m, n):
        ans = 0
        log_p_B = log_subset_probability(B, p_m)
        p_B = exp(log_p_B)
        if len(B) == 0:
            p_B = 1
            log_p_B = 0
        
        not_B = set()
        for attr in list(p_m.keys()):
            if not attr in B:
                not_B.add(attr)
        for k in range(n + 1):
            mult = 0
            mult_is_zero = False
            for attr in not_B:
                try:
                    mult += log(1 - ((p_m[attr]) ** k))
                except:
                    mult_is_zero = True
                    break
            if mult_is_zero:
                continue
            try:
                if p_B == 1 and n == k:
                    return exp(mult)
                else:
                    t = k * log_p_B + (n - k) * log((1 - p_B)) + mult
                    t = exp(t)
                    # print k, t
            except:
                t = 0
            nom = list(range(n - k + 1, n + 1))
            den = list(range(1, k + 1))
            if len(den) != len(nom):
                print("False")
            for i in range(len(nom)):
                t *= nom[i] / float(den[i])
            ans += t
        return ans

    def log_subset_probability(subset, p_m):
        ans = 0
        for attr in subset:
            try:
                ans += log(p_m[attr])
            except:
                pass
        return ans

    context = lattice.context
    n = len(context)
    p_m = {}
    for attr in context.attributes:
        m_ = 0
        for i in range(n):
            o = context.get_object_intent_by_index(i)
            if attr in o:
                m_ += 1
        p_m[attr] = m_ / float(n)
        
    probability = {}
    for concept in lattice:
        probability[concept] = get_intent_probability(concept.intent, p_m, n)
        
    return probability


## Rework to test in test file
# if __name__ == '__main__':
#     # Test code
#     from fcatng import ConceptLattice, Context
    
#     ct = [[True, False, False, True],\
#           [True, False, True, False],\
#           [False, True, True, False],\
#           [False, True, True, True]]
#     objs = [1, 2, 3, 4]
#     attrs = ['a', 'b', 'c', 'd']
#     c = Context(ct, objs, attrs)
#     cs = ConceptLattice(c)
#     ci = compute_probability(cs)
#     print(ci)

def compute_separation_index(lattice):
    context = lattice.context
    cross_index = {}
    
    for c in lattice:
        attrs = c.intent
        objs = c.extent
        square = len(attrs)*len(objs)
        crossed = 0
        for attr in attrs:
            crossed += len(context.get_attribute_extent(attr))
                 
        for obj in objs:
            crossed += len(context.get_object_intent(obj))
        crossed -= square
        if square == 0:
            cross_index[c] = 1.0
        else:
            cross_index[c] = square/crossed
            
    return cross_index


## Rework to testfile 
# if __name__ == '__main__':
#     from fcatng import ConceptLattice, Context
    
#     ct = [[True, False, False, True],\
#           [True, False, True, False],\
#           [False, True, True, False],\
#           [False, True, True, True]]
#     objs = [1, 2, 3, 4]
#     attrs = ['a', 'b', 'c', 'd']
#     c = Context(ct, objs, attrs)
#     cl = ConceptLattice(c)
#     ci = compute_separation_index(cl)
#     print(ci)




from copy import deepcopy

from fcatng import ConceptSystem


def compute_istability(lattice):
    """
    Examples
    ========

    >>> from fcatng import Context, ConceptLattice
    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> cl = ConceptLattice(c)
    >>> st = compute_estability(cl)
    >>> print st

    """
    concepts = ConceptSystem(lattice)
    count = {}
    subsets = {}
    stability = {}

    for concept in concepts:
        count[concept] = len([c for c in concepts if c.extent < concept.extent])
        subsets[concept] = 2 ** len(concept.extent)

    bottom_concepts = set([concepts.bottom_concept])
    while not len(concepts) == 0:
        bottom_concept = bottom_concepts.pop()
        stability[bottom_concept] = subsets[bottom_concept] / \
            (2 ** len(bottom_concept.extent))
        concepts.remove(bottom_concept)
        for c in concepts:
            if bottom_concept.intent > c.intent:
                subsets[c] -= subsets[bottom_concept]
                count[c] -= 1
                if count[c] == 0:
                    bottom_concepts.add(c)
    return stability


def compute_estability(lattice):
    concepts = ConceptSystem(lattice)
    count = {}
    subsets = {}
    stability = {}

    for concept in concepts:
        count[concept] = len([c for c in concepts if c.intent < concept.intent])
        subsets[concept] = 2 ** len(concept.intent)

    bottom_concepts = set([concepts.top_concept])
    while not len(concepts) == 0:
        bottom_concept = bottom_concepts.pop()
        stability[bottom_concept] = subsets[bottom_concept] / \
            (2 ** len(bottom_concept.intent))
        concepts.remove(bottom_concept)
        for c in concepts:
            if bottom_concept.intent < c.intent:
                subsets[c] -= subsets[bottom_concept]
                count[c] -= 1
                if count[c] == 0:
                    bottom_concepts.add(c)
    return stability




if __name__ == '__main__':
    # Test code
    from fcatng import ConceptLattice, Context
    from probability import compute_probability
    from stability import (compute_estability, compute_istability)
    from separation import compute_separation_index
    
    ct = [[True, False, False, True],\
          [True, False, True, False],\
          [False, True, True, False],\
          [False, True, True, True]]
    objs = ['1', '2', '3', '4']
    attrs = ['a', 'b', 'c', 'd']
    c = Context(ct, objs, attrs)
    cl = ConceptLattice(c)
#    compute_index(cl, compute_probability, "Probability")
#    cs = filter_concepts(cl, compute_probability, "abs", 4)
#    compute_index(cl, compute_separation_index, "Separation")
#    cs = filter_concepts(cl, compute_probability, "value", 0.5)
    # compute_index(cl, compute_istability, "Intensional Stability")
    # print cl
    cs = filter_concepts(cl, compute_istability, "abs", 2)
    print(cs)



# -*- coding: utf-8 -*-
"""
Contains base class for formal concept
"""

class Concept(object):
    """ 
    A formal concept, contains intent and extent 
    
    Examples
    ========

    Create a concept with extent=['Earth', 'Mars', 'Mercury', 'Venus']
    and intent=['Small size', 'Near to the sun'].
    
    >>> extent = ['Earth', 'Mars', 'Mercury', 'Venus']
    >>> intent = ['Small size', 'Near to the sun']
    >>> c = Concept(extent, intent)
    >>> 'Earth' in c.extent
    True
    >>> 'Pluto' in c.extent
    False
    >>> 'Small size' in c.intent
    True

    Print a concept.

    >>> print c
    (['Earth', 'Mars', 'Mercury', 'Venus'], ['Near to the sun', 'Small size'])

    """
    
    def __init__(self, extent, intent):
        """Initialize a concept with given extent and intent """
        self.extent = set(extent)
        self.intent = set(intent)
        self.meta = {}

    def __str__(self):
        """Return a string representation of a concept"""
        if len(self.intent) > 0:
            e = list(self.extent)
            e.sort()
        else:
            # TODO: Sometimes |intent| > 0, but extent is G.
            e = "G"
        if len(self.extent) > 0:
            i = list(self.intent)
            i.sort()
        else:
            # TODO: Sometimes |extent| > 0, but intent is M.
            i = "M"
        if len(list(self.meta.keys())) != 0:
            s = " meta: {0}".format(self.meta)
        else:
            s = ""
        return "({0}, {1}){2}".format(e, i, s)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8

import unittest
from . import closure_operators
import fcatng
from fcatng import compute_dg_basis
from fcatng.implication import Implication


class BasisTest(unittest.TestCase):
    def setUp(self):
        objects = ['Air Canada', 'Air New Zeland', 'All Nippon Airways',
                   'Ansett Australia', 'The Australian Airlines Group',
                   'British Midland', 'Lufthansa', 'Mexicana',
                   'Scandinavian Airlines', 'Singapore Airlines',
                   'Thai Airways International', 'United Airlines',
                   'VARIG']
        attributes = ['Latin America', 'Europe', 'Canada', 'Asia Pasific',
                      'Middle East', 'Africa', 'Mexico', 'Carribean',
                      'United States']
        table = [[True, True, True, True, True, False, True, True, True],
                 [False, True, False, True, False, False, False, False, True],
                 [False, True, False, True, False, False, False, False, True],
                 [False, False, False, True, False, False, False, False, False],
                 [False, True, True, True, True, True, False, False, True],
                 [False, True, False, False, False, False, False, False, False],
                 [True, True, True, True ,True, True, True, False, True],
                 [True, False, True, False, False, False, True, True, True],
                 [True, True, False, True, False, True, False, False, True],
                 [False, True, True, True, True, True, False, False, True],
                 [True, True, False, True, False, False, False, True, True],
                 [True, True, True, True, False, False, True, True, True],
                 [True, True, False, True, False, True, True, False, True]]
        self.cxt = fcatng.Context(table, objects, attributes)

class AirlinesTest(BasisTest):
    def test_attribute_implications(self):
        self.attribute_implications = [
            Implication(set([ 'Carribean', ]),set([ 'United States', 'Carribean', 'Latin America', ])),
            Implication(set([ 'Mexico', ]),set([ 'United States', 'Latin America', 'Mexico', ])),
            Implication(set([ 'Africa', ]),set([ 'Europe', 'United States', 'Asia Pasific', 'Africa', ])),
            Implication(set([ 'Middle East', ]),set([ 'Canada', 'Europe', 'Asia Pasific', 'United States', 'Middle East', ])),
            Implication(set([ 'United States', 'Asia Pasific', ]),set([ 'Europe', 'United States', 'Asia Pasific', ])),
            Implication(set([ 'Canada', ]),set([ 'Canada', 'United States', ])),
            Implication(set([ 'Europe', 'United States', ]),set([ 'Europe', 'United States', 'Asia Pasific', ])),
            Implication(set([ 'Europe', 'Asia Pasific', ]),set([ 'Europe', 'United States', 'Asia Pasific', ])),
            Implication(set([ 'Canada', 'Europe', 'Asia Pasific', 'Africa', 'United States', ]),set([ 'Canada', 'Europe', 'Asia Pasific', 'Africa', 'United States', 'Middle East', ])),
            Implication(set([ 'Latin America', ]),set([ 'United States', 'Latin America', ])),
            Implication(set([ 'United States', 'Carribean', 'Latin America', 'Mexico', ]),set([ 'Canada', 'United States', 'Carribean', 'Latin America', 'Mexico', ])),
            Implication(set([ 'Canada', 'United States', 'Latin America', ]),set([ 'Canada', 'United States', 'Latin America', 'Mexico', ])),
            Implication(set([ 'Europe', 'Asia Pasific', 'Africa', 'United States', 'Carribean', 'Latin America', ]),set([ 'Canada', 'Europe', 'Asia Pasific', 'Mexico', 'Africa', 'United States', 'Middle East', 'Carribean', 'Latin America', ])),
        ]
        
        imp_basis = compute_dg_basis(self.cxt)
        self.assertEqual(len(imp_basis), len(self.attribute_implications))
        for imp in self.attribute_implications:
	        self.assertTrue(imp in imp_basis)
	        
class RelativeBasisTest(unittest.TestCase):
    def setUp(self):
        ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
        objs = ['1', '2', '3', '4']
        attrs = ['a', 'b', 'c', 'd']
        self.cxt = fcatng.Context(ct, objs, attrs)
        self.basis = [Implication(set(['c', 'd']), set(['b']))]
    
    def test_relative_basis(self):
        simple_basis = compute_dg_basis(self.cxt)
        self.assertEqual(len(simple_basis), 3)
        relative_basis = compute_dg_basis(self.cxt, imp_basis=self.basis)
        self.assertEqual(len(set(relative_basis) & set(self.basis)), 0)
        self.assertEqual(len(relative_basis), 2)
        self.assertEqual(len(self.basis), 1)

class BasisTest2(unittest.TestCase):
    def setUp(self):
        ct = [[True, True, True, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
        objs = ['1', '2', '3', '4']
        attrs = ['a', 'b', 'c', 'd']
        self.cxt = fcatng.Context(ct, objs, attrs)
    
    def test_relative_basis(self):
        imp_basis=[Implication(set(), set(['c']))]
        relative_basis = compute_dg_basis(self.cxt, imp_basis=imp_basis)
        self.assertFalse(imp_basis[0] in relative_basis)

    def test_basis(self):
        simple_basis = compute_dg_basis(self.cxt)
        
        message = "Implication [{0}] occurs more than once"
        for imp1 in simple_basis:
            counter = 0
            for imp2 in simple_basis:
                if imp1 == imp2:
                    counter += 1
            self.assertEqual(counter, 1, message.format(imp1))

        
        


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-
"""
Holds class for context
"""
import copy

import fcatng.algorithms


class Context(object):
    """
    A Formal Context consists of two sets: *objects* and *attributes*
    and of a binary relation between them.

    As a data type we use a bit matrix.

    Examples
    ========

    Create a context.

    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> c[0][0]
    True
    >>> c[0][2]
    False
    >>> 1 in c.objects
    True
    >>> 'f' in c.attributes
    False
    >>> for o in c:
    ...     print o
    ...     break
    ...
    [True, False, False, True]
    >>> transposed_c = c.transpose()
    >>> for o in transposed_c:
    ...     print o
    ...
    [True, True, False, False]
    [False, False, True, True]
    [False, True, True, True]
    [True, False, False, True]

    Class emulates container type.

    >>> len(c)
    4
    >>> c[1]
    [True, False, True, False]

    Usage of examples.

    >>> c.get_object_intent_by_index(1)
    set(['a', 'c'])
    >>> for ex in c.examples():
    ...     print ex
    ...     break
    ...
    set(['a', 'd'])
    
    Implications basis
    
    >>> for imp in c.attribute_implications:
    ...     print imp
    ...
    c, d => b
    b => c
    a, c, b => d
    
    >>> for imp in c.object_implications:
    ...     print imp
    ...
    3 => 4
    2, 4 => 3
    1, 3, 4 => 2
    """

    def __init__(self, cross_table=[], objects=[], attributes=[]):
        """Create a context from cross table and list of objects, list
        of attributes
        
        cross_table - the list of bool lists
        objects - the list of objects
        attributes - the list of attributes 
        """
        if len(cross_table) != len(objects):
            raise ValueError("Number of objects (=%i) and number of cross table"
                   " rows(=%i) must agree" % (len(objects), len(cross_table)))
        elif (len(cross_table) != 0) and len(cross_table[0]) != len(attributes):
            raise ValueError("Number of attributes (=%i) and number of cross table"
                    " columns (=%i) must agree" % (len(attributes),
                        len(cross_table[0])))

        self._table = cross_table
        self._objects = objects
        self._attributes = attributes
        
    def __deepcopy__(self, memo):
        return Context(copy.deepcopy(self._table, memo),
                       self._objects[:],
                       self._attributes[:])

    def get_objects(self):
        return self._objects

    objects = property(get_objects)

    def get_attributes(self):
        return self._attributes

    attributes = property(get_attributes)
    
    def get_attribute_implications(self, 
                                   basis=fcatng.algorithms.compute_dg_basis,
                                   confirmed=[],
                                   cond=lambda x: True):
        return basis(self, imp_basis=confirmed, cond=cond)
        # if not self._attr_imp_basis or (confirmed != self._confirmed):
        #             self._attr_imp_basis = basis(self, imp_basis=confirmed)
        #             self._confirmed = confirmed
        #         return self._attr_imp_basis
    
    _attr_imp_basis = None
    _confirmed = None
    attribute_implications = property(get_attribute_implications)
    
    def get_object_implications(self, 
                                basis=fcatng.algorithms.compute_dg_basis,
                                confirmed=None):
        cxt = self.transpose()
        if not self._obj_imp_basis:
            self._obj_imp_basis = basis(cxt, imp_basis=confirmed)
        return self._obj_imp_basis
    
    _obj_imp_basis = None
    object_implications = property(get_object_implications)
        
    def examples(self):
        """Generator. Generate set of corresponding attributes
        for each row (object) of context
        """
        for obj in self._table:
            attrs_indexes = [i for i in range(len(obj)) if obj[i]]
            yield set([self.attributes[i] for i in attrs_indexes])
            
    def intents(self):
        return self.examples()

    def get_object_intent_by_index(self, i):
        """Return a set of corresponding attributes for row with index i"""
        # TODO: !!! Very inefficient. Avoid using
        attrs_indexes = [j for j in range(len(self._table[i])) if self._table[i][j]]
        return set([self.attributes[i] for i in attrs_indexes])
    
    def get_object_intent(self, o):
        index = self._objects.index(o)
        return self.get_object_intent_by_index(index)
    
    def get_attribute_extent_by_index(self, j):
        """Return a set of corresponding objects for column with index i"""
        objs_indexes = [i for i in range(len(self._table)) if self._table[i][j]]
        return set([self.objects[i] for i in objs_indexes])
    
    def get_attribute_extent(self, a):
        index = self._attributes.index(a)
        return self.get_attribute_extent_by_index(index)
        
    def get_value(self, o, a):
        io = self.objects.index(o)
        ia = self.attributes.index(a)
        return self[io][ia]
    
    def add_attribute(self, col, attr_name):
        """Add new attribute to context with given name"""
        for i in range(len(self._objects)):
            self._table[i].append(col[i])
        self._attributes.append(attr_name)

    def add_column(self, col, attr_name):
        """Deprecated. Use add_attribute."""
        print("Deprecated. Use add_attribute.")
        self.add_attribute(col, attr_name)

    def add_object(self, row, obj_name):
        """Add new object to context with given name"""
        self._table.append(row)
        self._objects.append(obj_name)
        
    def add_object_with_intent(self, intent, obj_name):
        self._attr_imp_basis = None
        self._objects.append(obj_name)
        row = [(attr in intent) for attr in self._attributes]
        self._table.append(row)
        
    def add_attribute_with_extent(self, extent, attr_name):
        col = [(obj in extent) for obj in self._objects]
        self.add_attribute(col, attr_name)
        
    def set_attribute_extent(self, extent, name):
        attr_index = self._attributes.index(name)
        for i in range(len(self._objects)):
            self._table[i][attr_index] = (self._objects[i] in extent)
            
    def set_object_intent(self, intent, name):
        obj_index = self._objects.index(name)
        for i in range(len(self._attributes)):
            self._table[obj_index][i] = (self._attributes[i] in intent)
        
    def delete_object(self, obj_index):
        del self._table[obj_index]
        del self._objects[obj_index]
        
    def delete_object_by_name(self, obj_name):
        self.delete_object(self.objects.index(obj_name))
    
    def delete_attribute(self, attr_index):
        for i in range(len(self._objects)):
            del self._table[i][attr_index]
        del self._attributes[attr_index]
        
    def delete_attribute_by_name(self, attr_name):
        self.delete_attribute(self.attributes.index(attr_name))
        
    def rename_object(self, old_name, name):
        self._objects[self._objects.index(old_name)] = name
        
    def rename_attribute(self, old_name, name):
        self._attributes[self._attributes.index(old_name)] = name
        
    def transpose(self):
        """Return new context with transposed cross-table"""
        new_objects = self._attributes[:]
        new_attributes = self._objects[:]
        new_cross_table = []
        for j in range(len(self._attributes)):
            line = []
            for i in range(len(self._objects)):
                line.append(self._table[i][j])
            new_cross_table.append(line)
        return Context(new_cross_table, new_objects, new_attributes)
        
    def extract_subcontext_filtered_by_attributes(self, attributes_names,
                                                    mode="and"):
        """Create a subcontext with such objects that have given attributes"""
        values = dict( [(attribute, True) for attribute in attributes_names] )
        object_names, subtable = \
                            self._extract_subtable_by_attribute_values(values, mode)
        return Context(subtable,
                       object_names,
                       self.attributes)
                            
    def extract_subcontext(self, attribute_names):
        """Create a subcontext with only indicated attributes"""
        return Context(self._extract_subtable(attribute_names),
                       self.objects,
                       attribute_names)
                                
    def _extract_subtable(self, attribute_names):
        self._check_attribute_names(attribute_names)
        attribute_indices = [self.attributes.index(a) for a in attribute_names] 
        table = []
        for i in range(len(self)):
            row = []
            for j in attribute_indices:
                row.append(self[i][j])
            table.append(row)
        
        return table
        
    def _extract_subtable_by_condition(self, condition):
        """Extract a subtable containing only rows that satisfy the condition.
        Return a list of object names and a subtable.
        
        Keyword arguments:
        condition(object_index) -- a function that takes an an object index and
            returns a Boolean value
        
        """
        indices = [i for i in range(len(self)) if condition(i)]
        return ([self.objects[i] for i in indices],
                [self._table[i] for i in indices])
                
    def _extract_subtable_by_attribute_values(self, values, 
                                                    mode="and"):
        """Extract a subtable containing only rows with certain column values.
        Return a list of object names and a subtable.
        
        Keyword arguments:
        values -- an attribute-value dictionary
        
        """
        self._check_attribute_names(list(values.keys()))
        if mode == "and":
            indices = [i for i in range(len(self)) if self._has_values(i, values)]
        elif mode == "or":
            indices = [i for i in range(len(self)) if self._has_at_least_one_value(i, values)]
        return ([self.objects[i] for i in indices],
                [self._table[i] for i in indices])
                
    def _has_values(self, i, values):
        """Test if ith object has attribute values as indicated.
        
        Keyword arguments:
        i -- an object index
        values -- an attribute-value dictionary
        
        """
        for a in values:
            j = self.attributes.index(a)
            v = values[a]
            if self[i][j] != v:
                return False
        return True
        
    def _has_at_least_one_value(self, i, values):
        """Test if ith object has at least one attribute value as in values.
                
        Keyword arguments:
        i -- an object index
        values -- an attribute-value dictionary
        
        """
        for a in values:
            j = self.attributes.index(a)
            v = values[a]
            if self[i][j] == v:
                return True
        return False
            
    def _check_attribute_names(self, attribute_names):
        if not set(attribute_names) <= set(self.attributes):
            wrong_attributes = ""
            for a in set(attribute_names) - set(self.attributes):
                wrong_attributes += "\t%s\n" % a
            raise ValueError("Wrong attribute names:\n%s" % wrong_attributes)    

    ############################
    # Emulating container type #
    ############################

    def __len__(self):
        return len(self._table)

    def __getitem__(self, key):
        return self._table[key]

    ############################
    
    def __repr__(self):
        output = ", ".join(self.attributes) + "\n"
        output += ", ".join(self.objects) + "\n"
        cross = {True : "X", False : "."}
        for i in range(len(self.objects)):
            output += ("".join([cross[b] for b in self[i]])) + "\n"
        return output

if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""Holds function that read context from tab separated txt file"""

import fcatng

def read_cxt(path):
    """Read context from path, which is .cxt file

    Format
    ======

    Example of .cxt file (tests/context.cxt):

    B

    4
    4

    Obj 1
    Obj 2
    Obj 3
    Obj 4
    a
    b
    c
    d
    X..X
    X.X.
    .XX.
    .XXX

    Examples
    ========

    Load example file from tests directory

    >>> c = read_cxt('tests/context.cxt')
    >>> len(c)
    4
    >>> len(c[0])
    4
    >>> for o in c:
    ...     print o
    ...
    [True, False, False, True]
    [True, False, True, False]
    [False, True, True, False]
    [False, True, True, True]
    >>> print c.objects
    ['Obj 1', 'Obj 2', 'Obj 3', 'Obj 4']
    >>> print c.attributes
    ['a', 'b', 'c', 'd']
    >>> c = read_cxt('tests/context.txt')
    Traceback (most recent call last):
        ...
    AssertionError: File is not valid cxt

    """
    input_file = open(path, "r")
    assert input_file.readline().strip() == "B",\
        "File is not valid cxt"
    input_file.readline() # Empty line
    number_of_objects = int(input_file.readline().strip())
    number_of_attributes = int(input_file.readline().strip())
    input_file.readline() # Empty line

    objects = [input_file.readline().strip() for i in range(number_of_objects)]
    attributes = [input_file.readline().strip() for i in range(number_of_attributes)]

    table = []
    for i in range(number_of_objects):
        line = [c=="X" for c in input_file.readline().strip()]
        table.append(line)

    input_file.close()

    return fcatng.Context(table, objects, attributes)


def write_cxt(context, path):
    output_file = open(path, "w")
    output_file.write("B\n\n")

    output_file.write(str(len(context.objects))+"\n")
    output_file.write(str(len(context.attributes))+"\n\n")

    for i in range(len(context.objects)):
        output_file.write(str(context.objects[i]))
        output_file.write("\n")

    for i in range(len(context.attributes)):
        output_file.write(str(context.attributes[i]))
        output_file.write("\n")

    cross = {True : "X", False : "."}
    for i in range(len(context.objects)):
        output_file.write("".join([cross[b] for b in context[i]]))
        output_file.write("\n")

    output_file.close()
    
def uwrite_cxt(context, path):
    output_file = open(path, "w")
    output_file.write("B\n\n")

    output_file.write(str(len(context.objects))+"\n")
    output_file.write(str(len(context.attributes))+"\n\n")

    for i in range(len(context.objects)):
        output_file.write(context.objects[i])
        output_file.write("\n")

    for i in range(len(context.attributes)):
        output_file.write(context.attributes[i])
        output_file.write("\n")

    cross = {True : "X", False : "."}
    for i in range(len(context.objects)):
        output_file.write("".join([cross[b] for b in context[i]]))
        output_file.write("\n")

    output_file.close()
    
def uread_cxt(path):
    input_file = open(path, "r")
    assert input_file.readline().strip() == "B",\
        "File is not valid cxt"
    input_file.readline() # Empty line
    number_of_objects = int(input_file.readline().strip())
    number_of_attributes = int(input_file.readline().strip())
    input_file.readline() # Empty line

    objects = [str(input_file.readline().strip(), "utf-8") for i in range(number_of_objects)]
    attributes = [str(input_file.readline().strip(), "utf-8") for i in range(number_of_attributes)]

    table = []
    for i in range(number_of_objects):
        line = [c=="X" for c in input_file.readline().strip()]
        table.append(line)

    input_file.close()

    return fcatng.Context(table, objects, attributes)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""Holds a function that scale a many-valued context to "one-valued" context"""

import fcatng

def scale_mvcontext(mvcontext, scales):
    """Scale many-valued context to one-valued. 
    Return fcatng.Context
    
    scales is a list of the scales(fcatng.Scale) applied
    to current many-valued context. Number of the scales
    in the list must agree with number of attributes in 
    many-valued context.
    
    Example
    =======
    
    >>> s = fcatng.read_cxt("tests/scale.cxt")
    >>> mvc = fcatng.read_mv_txt("../readwrite/tests/table.txt")
    >>> for o in mvc:
    ...     print o
    ...
    ['7', '6', '7']
    ['7', '2', '9']
    ['1', '3', '4']
    >>> c = scale_mvcontext(mvc, [s]*len(mvc.attributes))
    >>> for o in c:
    ...     print o
    ...
    [True, False, True, False, True, False]
    [True, False, False, False, True, False]
    [False, True, False, False, False, False]
    >>> print c.attributes
    ['attr1>5', 'attr1==1', 'attr2>5', 'attr2==1', 'attr3>5', 'attr3==1']
    >>> print c.objects
    ['g1', 'g2', 'g3']
    
    """
    derived_context = fcatng.Context([[] for i in range(len(mvcontext.objects))],
                                   mvcontext.objects, [])
    for attr_index in range(len(mvcontext.attributes)):
        scale = scales[attr_index]
        for col in range(len(scale.attributes)):
            derived_attr = [False for i in range(len(mvcontext.objects))]
            for row in range(len(scale.objects)):
                for obj in range(len(mvcontext.objects)):
                    if derived_attr[obj]:
                        continue
                    else:
                        try:
                            value = float(mvcontext[obj][attr_index])
                        except:
                            value = str(mvcontext[obj][attr_index])
                        if eval(scale.objects[row], {"value" : value}):
                            derived_attr[obj] = scale[row][col]
            new_attribute_name = ":".join([mvcontext.attributes[attr_index],
                                                  scale.attributes[col]])
            derived_context.add_column(derived_attr, new_attribute_name)                                    
    return derived_context
        

if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Holds functions that compute implication covers for a given context 
"""
import copy

from . import closure_operators
from fcatng.implication import Implication
import fcatng


def compute_implication_cover(cxt, close=closure_operators.closure):
    """
    Compute an implication cover for a given *cxt* using 
    an object-incremental algorithm
    """
    attributes = set(cxt.attributes)
    basis = [Implication(set(), attributes.copy())]
    i = 0
    for intent in cxt.examples():
        i += 1
        print(('object ', i))
#        print_basis(basis)
#        print 'Adding ', intent
#        raw_input()
        basis = updated_basis(intent, basis, attributes)
        print((len(basis), 'implications'))
    return basis


def print_basis(basis):
    print('***')
    for imp in basis:
        print(imp)
    print('+++')
    

def is_redundant(imp, basis, close=closure_operators.simple_closure):
    return imp.conclusion <= close(imp.premise, basis)
    
    
def is_new(imp, implications):
    if imp.conclusion <= imp.premise:
        return False
    for i in implications:
        if is_subsumed_simply(imp, i):
            return False
        elif is_subsumed(imp, i):
            raise 'ALERT: %s is sumbsumed by %s' %(imp, i)
        elif is_subsumed(i, imp):
            raise 'ALERT: %s sumbsumes %s' % (imp, i)
    return True
    

def is_subsumed_simply(imp, by_imp):
    return by_imp.premise <= imp.premise and imp.conclusion <= by_imp.conclusion


def is_subsumed(imp, by_imp):
    return (by_imp.premise <= imp.premise and
            imp.conclusion <= (by_imp.conclusion | imp.premise))
    

def remove_subsumed_plus(imp, implications):
    if imp.conclusion <= imp.premise:
        return False
    for i in implications[:]:
        if i.premise == imp.premise:
            i.conclusion.update(imp.conclusion)
            return False
        elif i.premise <= imp.premise:
            if imp.conclusion <= i.conclusion:
                return False
            else:
                imp.conclusion.update(i.conclusion)
        elif imp.premise <= i.premise:
            if i.conclusion <= imp.conclusion:
                implications.remove(i)
            else:
                i.conclusion.update(imp.conclusion)
    return True


def remove_subsumed(imp, implications):
    if imp.conclusion <= imp.premise:
        return False
    can_be_subsumed = True
    for i in implications[:]:
        if can_be_subsumed and i.premise <= imp.premise:
            if imp.conclusion <= i.conclusion:
                return False
            else:
                imp.conclusion.update(i.conclusion)
        elif is_subsumed_simply(i, imp):
            implications.remove(i)
            can_be_subsumed = False
    return True
    

def remove_subsumed_simply(imp, implications):
    if imp.conclusion <= imp.premise:
        return False
    can_be_subsumed = True
    for i in implications[:]:
        if can_be_subsumed and is_subsumed_simply(imp, i):
            return False
        elif is_subsumed_simply(i, imp):
            implications.remove(i)
            can_be_subsumed = False
    return True


def add_smartly(imp, new_valid, valid):
    if is_new(imp, valid) and remove_subsumed_simply(imp, new_valid):
        new_valid.append(imp)


def updated_basis(intent, basis, attributes):
    valid = []
    invalid = []
    for imp in basis:
        if not imp.premise <= intent or imp.conclusion <= intent:
            valid.append(imp)
        else:
            invalid.append(imp)
            
    new_valid = []
    for imp in invalid:
        new_imp = Implication(imp.premise, imp.conclusion & intent)
        add_smartly(new_imp, new_valid, valid)
        
    valid += new_valid
    new_valid = []
    for imp in invalid:
        for a in attributes - intent:
            aset = set([a])
            new_imp = Implication(imp.premise | aset, imp.conclusion | aset)
            if (remove_subsumed(new_imp, valid) and
                remove_subsumed_plus(new_imp, new_valid)):
               new_valid.append(new_imp)
    return valid + new_valid
                    

def minimize(cover, close=closure_operators.simple_closure):
#    i = 0
#    for imp in cover:
#        i += 1
#        print 'maximizing conclusion ', i
#        imp._conclusion = close(imp.premise | imp.conclusion, cover)
    i = 0
    for imp in cover[:]:
        i += 1
        print(('maximizing premise ', i))
        cover.remove(imp)
        imp._premise = close(imp.premise, cover)
        if not imp.conclusion <= imp.premise:
            cover.append(imp)
        print((len(cover), 'implications'))


if __name__ == "__main__":    
    objects = ['Air Canada', 'Air New Zeland', 'All Nippon Airways',
               'Ansett Australia', 'The Australian Airlines Group',
               'British Midland', 'Lufthansa', 'Mexicana',
               'Scandinavian Airlines', 'Singapore Airlines',
               'Thai Airways International', 'United Airlines',
               'VARIG']
    attributes = ['Latin America', 'Europe', 'Canada', 'Asia Pasific',
                  'Middle East', 'Africa', 'Mexico', 'Carribean',
                  'United States']
    table = [[True, True, True, True, True, False, True, True, True],
             [False, True, False, True, False, False, False, False, True],
             [False, True, False, True, False, False, False, False, True],
             [False, False, False, True, False, False, False, False, False],
             [False, True, True, True, True, True, False, False, True],
             [False, True, False, False, False, False, False, False, False],
             [True, True, True, True ,True, True, True, False, True],
             [True, False, True, False, False, False, True, True, True],
             [True, True, False, True, False, True, False, False, True],
             [False, True, True, True, True, True, False, False, True],
             [True, True, False, True, False, False, False, True, True],
             [True, True, True, True, False, False, True, True, True],
             [True, True, False, True, False, True, True, False, True]]
    cxt = fcatng.Context(table, objects, attributes)
    
    imp_basis = compute_implication_cover(cxt, closure_operators.closure)
    print_basis(imp_basis)
    minimize(imp_basis)
    print((len(imp_basis)))
    for imp in imp_basis:
        print(imp)
        
    print('***')
    
    objects = [1, 2, 3, 4]
    attributes = ['a', 'b', 'c', 'd']
    table = [[True, False, True, True],
             [True, False, True, False],
             [False, True, True, False],
             [False, True, False, True]]
    cxt = fcatng.Context(table, objects, attributes)
    
    imp_basis = compute_implication_cover(cxt, closure_operators.closure)
    print_basis(imp_basis)
    minimize(imp_basis)
    print((len(imp_basis)))
    for imp in imp_basis:
        print(imp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import copy

import fcatng

class ExplorationException(Exception):
    pass

class FalseCounterexample(ExplorationException):
    pass

class IllegalContextModification(ExplorationException):

    def __str__(self):
        return "Attempted context modification conflicts with existing background knowledge"

class ExplorationContext(object):
    """Implements lightweight interface to formal context"""
    _cxt = None
    _basis = None

    def get_objects(self):
        return copy(self._cxt.get_objects())

    objects = property(get_objects)

    def get_attributes(self):
        return copy(self._cxt.get_attributes())

    attributes = property(get_attributes)

    def __init__(self, cxt):
        self._cxt = cxt

    def set_basis(self, basis):
        self._basis = basis

    def get_intent(self, object):
        return copy(self._cxt.get_object_intent(object))

    def get_extent(self, attribute):
        return copy(self._cxt.get_attribute_extent(attribute))

    def add_object(self, object, intent):
        if self._basis.respect_background_knowledge(intent):
            self._cxt.add_object_with_intent(intent, object)
        else:
            raise IllegalContextModification()

class Basis(object):
    _cxt = None
    _background_implications = None
    _basis = None

    def __init__(self, cxt):
        self._cxt = cxt
        self._background_implications = []

    def get_dg_basis(self):
        return fcatng.compute_dg_basis(self._cxt._cxt, imp_basis=self._background_implications)

    def get_background_implications(self):
        return copy(self._background_implications)

    def add_background_implication(self, imp):
        self._background_implications.append(imp)

    def respect_background_knowledge(self, intent):
        for imp in self._background_implications:
            if not imp.is_respected(intent):
                return False
        return True

class ExplorationSession(object):
    _cxt = None
    _basis = None

    def __init__(self, cxt):
        self._cxt = cxt
        self._basis = Basis(cxt)
        self._cxt.set_basis(self._basis)

    def get_context(self):
        return self._cxt

    context = property(get_context)

    def get_accepted_implications(self):
        return self._basis.get_background_implications()

    def get_candidates(self):
        return self._basis.get_dg_basis()

    def accept_implication(self, imp):
        self._basis.add_background_implication(imp)

    def reject_implication(self, imp, object, intent):
        if not imp.is_respected(intent):
            self._cxt.add_object(object, intent)
        else:
            raise FalseCounterexample()

class Exploration(object):
    _cxt = None

    def __init__(self, cxt):
        self._cxt = copy(cxt)

    def set_context(self, cxt):
        self._cxt = copy(cxt)

    def create_session(self):
        return ExplorationSession(self._cxt)




#  conceptual_exploration.py
#  
#
#  Created by object on 3/13/11.
#  Copyright (c) 2011 __MyCompanyName__. All rights reserved.
#

class WrongCounterExample(ExplorationException):
    pass


class Exploration(object):

    def __init__(self, db, expert):
        self.db = db
        self.expert = expert
            
    def confirm_implication(self, impl):
        self.db.add_implication(impl)
        
    def reject_implication(self, impl):
        counterexample = self.expert.provide_counterexample(impl)
        if self.db.respects(counterexample, impl):
            raise WrongCounterExample()
        self.db.add_example(counterexample)
        
#    def on_context_update(self):
        



def write_fimi(context, path):
    output_file = open(path, "w")

    for o in context.objects:
        intent = context.get_object_intent(o)
        for a in intent:
            output_file.write('%s ' % a)
        output_file.write('\n')

    output_file.close()



# -*- coding: utf-8 -*-
"""
Holds scale class for conceptual scaling
"""

from .context import Context

class Scale(Context):
    """
    Scale inherited from the Context used for conceptual scaling.
    
    Examples
    ========
    
    >>> ct = [[True, False],\
              [False, True]]
    >>> objs = ['value>7', 'value<2']
    >>> attrs = ['>7', '<2']
    >>> c = Context(ct, objs, attrs)
    >>> s = Scale(c)
    
    """

    def __init__(self, context):
        """Constructor"""
        super(Scale, self).__init__(context._table,
                                    context._objects,
                                    context._attributes)


class NominalScale(Scale):
    def __init__(self, attribute, mvcontext):
        """Generate a nominal scale for an attribute based on its values in the
        input many-valued context.
        
        """
        if type(attribute) == str:
            attribute = mvcontext.attributes.index(attribute)
        values = list(set([mvcontext[i][attribute]
                           for i in range(len(mvcontext))]))
        objects = ["value == %s" % v for v in values]
        table = []
        for i in range(len(objects)):
            row = [False] * len(values)
            row[i] = True
            table.append(row)
        super(NominalScale, self).__init__(Context(table, objects, values))


class OrdinalScale(Scale):
    def __init__(self, min, max, mode="leq"):
        """Generate an ordinal scale for values between min and max.
        
        """
        values = list(range(min, max + 1))
        objects = ["value == %s" % v for v in values]
        table = []
        if mode == "geq":
            attributes = [">= %s" % v for v in values[1:]]
            for i in range(len(objects)):
                table.append([True] * i + [False] * (len(attributes) - i))
        else:
            attributes = ["<= %s" % v for v in values[:-1]]
            for i in range(len(objects)):
                table.append([False] * i + [True] * (len(attributes) - i))
        super(OrdinalScale, self).__init__(Context(table, objects, attributes))

        
if __name__ == "__main__":
    import doctest
    doctest.testmod()
        


# -*- coding: utf-8 -*-
"""
Holds a function that computes Duquenne-Guigues basis for a given context 
"""
import copy

from . import closure_operators
from fcatng.implication import Implication
import fcatng

def compute_dg_basis(cxt,
                     close=closure_operators.lin_closure,
                     imp_basis=[],
                     cond=lambda x: True):
    """
    Compute Duquenne-Guigues basis for a given *cxt* using 
    optimized Ganter algorithm
    """
    aclose = lambda attributes: closure_operators.aclosure(attributes, cxt)
    return generalized_compute_dg_basis(cxt.attributes, 
                                        aclose,
                                        close=close,
                                        imp_basis=imp_basis,
                                        cond=cond)


def compute_partial_dg_basis(pcxt,
                             close=closure_operators.simple_closure,
                             imp_basis=[],
                             cond=lambda x: True):
    """
    Compute Duquenne-Guigues basis for a given partial context *pcxt* using
    optimized Ganter algorithm
    """
    return generalized_compute_dg_basis(pcxt.attributes,
                                        pcxt.xq_aclosure,
                                        close=close,
                                        imp_basis=imp_basis,
                                        cond=cond)


def generalized_compute_dg_basis(attributes,
                                 aclose,
                                 close=closure_operators.simple_closure,
                                 imp_basis=[],
                                 cond=lambda x: True):
    """Compute the Duquenne-Guigues basis using optimized Ganter's algorithm.
    
    *aclose* is a closure operator on the set of attributes.
    We need this to implement the exploration
    algorithm with partially given examples.
    
    """
    relative_basis = []
    
    a = close(set(), imp_basis)
    i = len(attributes)
    
    while len(a) < len(attributes):
        a_closed = set(aclose(a))
        if a != a_closed and cond(a):
            relative_basis.append(Implication(a.copy(), a_closed.copy()))
        if (a_closed - a) & set(attributes[: i]):
            a -= set(attributes[i :])
        else:
            if len(a_closed) == len(attributes):
                return relative_basis
            a = a_closed
            i = len(attributes)
        for j in range(i - 1, -1, -1):
            m = attributes[j]
            if m in a:
                a.remove(m)
            else:
                b = close(a | set([m]), relative_basis + imp_basis)
                if not (b - a) & set(attributes[: j]):
                    a = b
                    i = j
                    break

    return relative_basis

if __name__ == "__main__":    
    # objects = ['Air Canada', 'Air New Zeland', 'All Nippon Airways',
    #            'Ansett Australia', 'The Australian Airlines Group',
    #            'British Midland', 'Lufthansa', 'Mexicana',
    #            'Scandinavian Airlines', 'Singapore Airlines',
    #            'Thai Airways International', 'United Airlines',
    #            'VARIG']
    # attributes = ['Latin America', 'Europe', 'Canada', 'Asia Pasific',
    #               'Middle East', 'Africa', 'Mexico', 'Carribean',
    #               'United States']
    # table = [[True, True, True, True, True, False, True, True, True],
    #          [False, True, False, True, False, False, False, False, True],
    #          [False, True, False, True, False, False, False, False, True],
    #          [False, False, False, True, False, False, False, False, False],
    #          [False, True, True, True, True, True, False, False, True],
    #          [False, True, False, False, False, False, False, False, False],
    #          [True, True, True, True ,True, True, True, False, True],
    #          [True, False, True, False, False, False, True, True, True],
    #          [True, True, False, True, False, True, False, False, True],
    #          [False, True, True, True, True, True, False, False, True],
    #          [True, True, False, True, False, False, False, True, True],
    #          [True, True, True, True, False, False, True, True, True],
    #          [True, True, False, True, False, True, True, False, True]]
    # cxt = fcatng.Context(table, objects, attributes)
    ct = [[True]]
    objs = ['1']
    attrs = ['a']
    cxt = fcatng.Context(ct, objs, attrs)

    imp_basis = compute_dg_basis(cxt, imp_basis=[Implication(set(), set(['a']))])

    for imp in imp_basis:
        print(imp)



# -*- coding: utf-8 -*-
"""
Contains class for implications
"""

class Implication(object):
    """
    An Implication consists of two sets: *premise* and *conclusion*
    
    Examples
    ========
    
    >>> imp = Implication(set(('a', 'b',)), set(('c',)))
    >>> imp
    a, b => c
    >>> print imp
    a, b => c
    >>> imp.is_respected(set(('a', 'b',)))
    False
    >>> imp.is_respected(set(('a', 'b', 'd')))
    False
    >>> imp.is_respected(set(('a', 'b', 'c',)))
    True
    >>> imp.is_respected(set(('a', 'c',)))
    True
    >>> imp.is_respected(set(('b')))
    True
    >>> imp.is_respected(set(('c')))
    True
    """

    def __init__(self, premise = set(), conclusion = set()):
        """
        Create implication from two sets of attributes
        """
        self._premise = premise
        self._conclusion = conclusion
        
    def __deepcopy__(self, memo):
        return Implication(self._premise.copy(), self._conclusion.copy())

    def get_premise(self):
        """
        Return premise of implication
        """
        return self._premise
        
    def get_conclusion(self):
        """
        Return conclusion of implication
        """
        return self._conclusion

    def get_reduced_conclusion(self):
        return self._conclusion - self._premise
    
    premise = property(get_premise)
    conclusion = property(get_reduced_conclusion)
        
    def __repr__(self):
        try:
            premise = ", ".join([element for element in self._premise])
            short_conclusion = self._conclusion - self._premise
            conclusion = ", ".join([element for element in short_conclusion])
        except:
            premise = ", ".join([str(element) for element in self._premise])
            short_conclusion = self._conclusion - self._premise
            conclusion = ", ".join([str(element) for element in short_conclusion])
        return " => ".join((premise, conclusion,))
        
    def __unicode__(self):
        return self.__repr__()

    def __cmp__(self, other):
        if ((self._premise == other.premise) and 
            (self._conclusion == other.conclusion)):
            return 0
        else:
            return -1
            
    def is_respected(self, some_set):
        """Checks whether *some_set* respects an implication or not"""
        # if some_set contains every element from premise and not every
        # element from conclusion then it doesn't respect an implication
        # TODO: refactor
        if type(some_set) == set:
            return self.conclusion <= some_set or not self.premise <= some_set
        else:
            # Assume a partial example
            return (self.conclusion <= some_set[1] or
                    not self.premise <= some_set[0])
        
if __name__ == "__main__":
    import doctest
    doctest.testmod()


import copy

import fcatng

class ExplorationException(Exception):
    pass
    
class WrongCounterExample(ExplorationException):
    pass
    
class BasisConflict(ExplorationException):
    
    def __init__(self, set_):
        self.set = set_
        
    def __str__(self):
        return "{0} conflict with confirmed implications".format(self.set)
    
class NotCounterExamplePremise(WrongCounterExample):
    
    def __init__(self, set_, implication):
        self.set = set_
        self.implication = implication
        
    def __str__(self):
        output = "{0} is not counter example to {1}. ".format(self.set, 
                                                            self.implication)
        output += "Counter example doesn't contain all elements from premise"
        return output
        
class NotCounterExampleConclusion(WrongCounterExample):

    def __init__(self, set_, implication):
        self.set = set_
        self.implication = implication

    def __str__(self):
        output = "{0} is not counter example to {1}. ".format(self.set, 
                                                            self.implication)
        output += "Counter example intent contains all elements from "\
                  "conclusion"
        return output

class BasicExploration(object):
    """
    Exploration class for basic exploration algorithm
    """
    
    context = None
    attribute_implications = None
    object_implications = None
    confirmed_attribute_implications = None
    confirmed_object_implications = None
    
    def __init__(self, initial_cxt):
        """Exploration starts with some initial context - *initial_cxt*"""
        self.context = copy.deepcopy(initial_cxt)
        # After initializng of context we need to compute initial implications
        self._init_implications()
        
    def __repr__(self):
        output = ""
        output += "Context:\n"
        output += "========\n"
        output += str(self.context) + "\n"
        output += "Attribute implications:\n"
        output += "=======================\n"
        for imp in self.attribute_implications:
            output += str(imp) + "\n"
        output += "Confirmed attribute implications:\n"
        output += "=================================\n"
        for imp in self.confirmed_attribute_implications:
            output += str(imp) + "\n"
        output += "Object implications:\n"
        output += "====================\n"
        for imp in self.object_implications:
            output += str(imp) + "\n"
        output += "Confirmed object implications:\n"
        output += "==============================\n"
        for imp in self.confirmed_object_implications:
            output += str(imp) + "\n"
        return output
        
    def _init_implications(self):
        """Compute stem base for initial context"""
        self.attribute_implications = fcatng.compute_dg_basis(self.context)
        transposed_cxt = self.context.transpose()
        self.object_implications = fcatng.compute_dg_basis(transposed_cxt)
        self.confirmed_attribute_implications = []
        self.confirmed_object_implications = []
        
    def recompute_basis(self):
        basis = self.confirmed_attribute_implications
        new_implications = fcatng.compute_dg_basis(self.context, basis)
        self.attribute_implications = []
        for imp in new_implications:
            if imp not in basis:
                self.attribute_implications.append(imp)
                
        basis = self.confirmed_object_implications
        transposed_cxt = self.context.transpose()
        new_implications = fcatng.compute_dg_basis(transposed_cxt, basis)
        self.object_implications = []
        for imp in new_implications:
            if imp not in basis:
                self.object_implications.append(imp)
        
    def confirm_attribute_implication(self, imp_index):
        imp = self.attribute_implications[imp_index]
        self.confirmed_attribute_implications.append(imp)
        del self.attribute_implications[imp_index]
        
    def confirm_object_implication(self, imp_index):
        imp = self.object_implications[imp_index]
        self.confirmed_object_implications.append(imp)
        del self.object_implications[imp_index]
        
    def counter_example_for_attr_implication(self, name, intent, imp_index):
        implication = self.attribute_implications[imp_index]
        premise = implication.premise
        conclusion = implication.conclusion
        
        # first we need to check that it is actually the counter example
        # counter example intent should contain all attributes from premise
        if (premise & intent) != premise:
            raise NotCounterExamplePremise(intent, implication)

        # Counterexample intent should contain NOT all attributes from 
        # conclusion
        if (conclusion & intent) == conclusion:
            raise NotCounterExampleConclusion(intent, implication)
        
        if not self.check_intent_for_conflicts(intent):
            raise BasisConflict(intent)    
            
        # if counter example is correct, we add new object to context
        # and recompute stem base with confirmed implications as basis
        self.context.add_object_with_intent(intent, name)
        self.recompute_basis()
            
    def counter_example_for_obj_implication(self, name, extent, imp_index):
        implication = self.object_implications[imp_index]
        premise = implication.premise
        conclusion = implication.conclusion

        # first we need to check that it is actually the counter example
        # counter example intent should contain all attributes from premise
        if (premise & extent) != premise:
            raise NotCounterExamplePremise(extent, implication)

        # Counterexample intent should contain NOT all attributes from 
        # conclusion
        if (conclusion & extent) == conclusion:
            raise NotCounterExampleConclusion(extent, implication)
            
        if not self.check_extent_for_conflicts(extent):
            raise BasisConflict(extent)
            
        # if counter example is correct, we add new attribute to context
        # and recompute stem base with confirmed implications as basis
        self.context.add_attribute_with_extent(extent, name)
        self.recompute_basis()    
        
    def check_extent_for_conflicts(self, extent):
        """
        Checks new attribute with *extent* for conflicts with confirmed
        object implications. Return True if all is ok.
        """
        for imp in self.confirmed_object_implications:
            if (imp.premise & extent) != imp.premise:
                continue
            if (imp.conclusion & extent) == imp.conclusion:
                continue
            return False

        return True
        
    def check_intent_for_conflicts(self, intent):
        """
        Checks new object with *intent* for conflicts with confirmed
        attribute implications. Return True if all is ok.
        """
        for imp in self.confirmed_attribute_implications:
            if (imp.premise & intent) != imp.premise:
                continue
            if (imp.conclusion & intent) == imp.conclusion:
                continue
            return False

        return True
        
    def add_object(self, intent, name):
        if not check_intent_for_conflicts(intent):
            raise BasisConflict(intent)
        else:
            self.context.add_object_with_intent(intent, name)
            self.recompute_basis()
            
    def add_attribute(self, extent, name):
        if not check_extent_for_conflicts(extent):
            raise BasisConflict(extent)
        else:
            self.context.add_attribute_with_extent(extent, name)
            self.recompute_basis()
            
    def edit_attribute(self, new_extent, name):
        if not check_extent_for_conflicts(extent):
            raise BasisConflict(extent)
        else:
            self.context.set_attribute_extent(extent, name)
            self.recompute_basis()
            
    def edit_object(self, new_intent, name):
        if not check_intent_for_conflicts(intent):
            raise BasisConflict(intent)
        else:
            self.context.set_object_intent(intent, name)
            self.recompute_basis()

if __name__ == "__main__":    
    table = [[True, False, False, True],\
             [True, False, True, False],\
             [False, True, True, False],\
             [False, True, True, True]]
    objs = ['1', '2', '3', '4']
    attrs = ['a', 'b', 'c', 'd']
    cxt = fcatng.Context(table, objs, attrs)
    exp = BasicExploration(cxt)
    print(exp)
    exp.confirm_object_implication(1)
    exp.counter_example_for_obj_implication('test', set(['3']), 0)
    print(exp)



# -*- coding: utf-8 -*-
"""Holds implementation of Norris' algorithm"""

from copy import copy
from fcatng import Concept, ConceptSystem

def derivation(context, extent=0, intent=0):
    """Return result of derivation operator applied to intent or extent.
    If extent==0 applied to intent and
    if intent==0 applied to extent

    TODO: May be move to context module
    TODO: A lot of lines, rewrite shorter
    TODO: !!! Seems to be useless in this module

    Examples
    ========
    >>> from fcatng import Context
    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> derivation(c, extent=set([1, 2]))
    set(['a'])
    >>> derivation(c, intent=set(['b', 'c']))
    set([3, 4])

    """
    result = set()
    if intent != 0: # if we apply to intent
        if intent == set(context.attributes):
            return set()
        elif len(intent) == 0:
            return set(context.objects)
        else:
            first_attr = True
            for attr in intent:
                objects = set()
                index = context.attributes.index(attr)
                for j in range(len(context)):
                    if context[j][index]: # if object has attribute
                        objects.add(context.objects[j])
                # if first_attr - result is empty set and we cant intersect it
                # with objects
                if first_attr:
                    result = copy(objects)
                    first_attr = False
                else:
                    result &= objects
                if len(result) == 0:
                    # if result become empty there are no meaning to complete
                    # computation
                    return result
            return result
    else: # extent
        if extent == set(context.objects):
            return set()
        elif len(extent) == 0:
            return set(context.attributes)
        else:
            first_attr = True
            for obj in extent:
                attrs = set()
                index = context.objects.index(obj)
                for j in range(len(context[index])):
                    if context[index][j]: # if attribute has object
                        attrs.add(context.attributes[j])
                # if first_attr - result is empty set and we cant intersect it
                # with attrs 
                if first_attr:
                    result = copy(attrs)
                    first_attr = False
                else:
                    result &= attrs
                if len(result) == 0:
                    # if result become empty there are no meaning to complete
                    # computation
                    return result
            return result
                        

def norris(context):
    """Build all concepts of a context
    
    Based on the Norris' algorithm for computing the maximal rectangles
    in a binary relation.

    Returns the ConceptSystem instance with a full set of concepts

    Examples
    ========

    >>> from fcatng import Context, ConceptLattice
    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> cl = ConceptLattice(c, builder=norris)
    >>> print cl
    ([], M)
    ([1], ['a', 'd'])
    ([2], ['a', 'c'])
    ([1, 2], ['a'])
    ([3, 4], ['b', 'c'])
    ([2, 3, 4], ['c'])
    (G, [])
    ([4], ['b', 'c', 'd'])
    ([1, 4], ['d'])

    """
    # To be more efficient we store intent (as Python set) of every 
    # object to the list
    # TODO: Move to Context class?
    examples = []
    for ex in context.examples():
        examples.append(ex)
    
    cs = [Concept([], context.attributes)]
    for i in range(len(context)):
        # TODO:
        cs_for_loop = cs[:]
        for c in cs_for_loop:
            if c.intent.issubset(examples[i]):
                c.extent.add(context.objects[i])
            else:
                new_intent = c.intent & examples[i]
                new = True
                for j in range(i):
                    if new_intent.issubset(examples[j]) and\
                       context.objects[j] not in c.extent:
                        new = False
                        break
                if new:
                    cs.append(Concept(set([context.objects[i]]) | c.extent,
                        new_intent))
    return (cs, compute_covering_relation(cs))


def compute_covering_relation(cs):
        """Computes covering relation for a given concept system.

        Returns a dictionary containing sets of parents for each concept.

        Examples
        ========

        """
        parents = dict([(c, set()) for c in cs])

        for i in range(len(cs)):
            for j in range(len(cs)):
                if cs[i].intent < cs[j].intent:
                    parents[cs[j]].add(cs[i])
                    for k in range(len(cs)):
                        if cs[i].intent < cs[k].intent and\
                           cs[k].intent < cs[j].intent:
                                parents[cs[j]].remove(cs[i])
                                break
        return parents


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from fcatng.algorithms.exploration2.exploration import *

def is_prime(n):
    for m in xrange(n - 1, 1, -1):
        if n % m == 0:
            return False

    return True

def is_factorial(n):
    f = 1
    for m in xrange(1, n + 1):
        f *= m
        if f == n:
            return True
        if f > n:
            return False



class CommandLineExploration(Exploration):
    def __init__(self):
        self.d = {
            "even" : lambda n: n % 2 == 0,
            "odd" : lambda n: n % 2 == 1,
            "divided_by_three" : lambda n: n % 3 == 0,
            "prime" : is_prime,
            "factorial" : is_factorial
        }

        cxt = fcatng.Context(attributes=list(self.d.keys()))

        super(CommandLineExploration, self).__init__(ExplorationContext(cxt))
        self._session = self.create_session()

    def is_valid(self, imp):
        print "{0}".format(imp)
        return input('Is the following implication valid? Enter "True" or "False":\n'.format(imp))

    def ask_for_counterexample(self):
        return input('Provide counterexample:')

    def get_intent(self, number):
        return {attr for attr in self._cxt.attributes if self.d[attr](number) }

    def explore(self):
        while self._session.get_candidates():
            imp = self._session.get_candidates()[0]
            if self.is_valid(imp):
                self._session.accept_implication(imp)
            else:
                counterexample = self.ask_for_counterexample()
                intent = self.get_intent(counterexample)
                print intent
                self._session.reject_implication(imp, counterexample, intent)

if __name__ == "__main__":
    CommandLineExploration().explore()







#!/usr/bin/env python
# encoding: utf-8
"""
explorationtest.py
"""

import unittest
from .exploration import (BasicExploration, NotCounterExamplePremise, 
                        NotCounterExampleConclusion)

import fcatng


class ExplorationTest(unittest.TestCase):
    def setUp(self):
        table = [[True, False, False, True],\
                 [True, False, True, False],\
                 [False, True, True, False],\
                 [False, True, True, True]]
        objs = ['1', '2', '3', '4']
        attrs = ['a', 'b', 'c', 'd']
        self.exp = BasicExploration(fcatng.Context(table, objs, attrs))
        
class TestNotEmptyConfirmedList(ExplorationTest):
        
    def test_counter_examples_for_object_implication(self):
        self.exp.confirm_object_implication(1)
        confirmed_implication = self.exp.confirmed_object_implications[0]

        number_of_attrs_before = len(self.exp.context.attributes)
        self.exp.counter_example_for_obj_implication(
                                    'test', set(['3']), 0)
        number_of_attrs_after = len(self.exp.context.attributes)
        self.assertEqual(number_of_attrs_after - number_of_attrs_before, 
                            1, "Number of attributes isn't increased")
                            
        self.assertFalse(confirmed_implication in self.exp.object_implications)
                            
    def test_counter_examples_for_attribute_implication(self):
        self.exp.confirm_attribute_implication(1)
        confirmed_implication = self.exp.confirmed_attribute_implications[0]
        
        number_of_objects_before = len(self.exp.context.objects)
        self.exp.counter_example_for_attr_implication(
                                    'test', set(['c', 'd']), 0)
        number_of_objects_after = len(self.exp.context.objects)
        self.assertEqual(number_of_objects_after - number_of_objects_before, 
                            1, "Number of objects isn't increased")
                            
        self.assertFalse(confirmed_implication in self.exp.attribute_implications)
           
class TestExplorationInitialization(ExplorationTest):
    """docstring for TestExplorationInitialization"""
    def runTest(self):
        self.assertEqual(len(self.exp.attribute_implications), 3,
                         "Wrong number of attribute implications")
        self.assertEqual(len(self.exp.object_implications), 3,
                         "Wrong number of object implications")
                         
class TestImplicationsConfirming(ExplorationTest):
    
    def test_object_implication_confirm(self):
        self.exp.confirm_attribute_implication(0)
        self.assertEqual(len(self.exp.attribute_implications), 2,
                         "Wrong number of attribute implications")
        self.assertEqual(len(self.exp.confirmed_attribute_implications), 1,
                         "Wrong number of confirmed attribute implications")
                         
        self.exp.confirm_object_implication(0)
        self.assertEqual(len(self.exp.object_implications), 2,
                         "Wrong number of object implications")
        self.assertEqual(len(self.exp.confirmed_object_implications), 1,
                         "Wrong number of object attribute implications")
        
class TestCounterExample(ExplorationTest):
    
    def test_providing_wrong_counter_examples(self):
        self.assertRaises(
            NotCounterExamplePremise,
            self.exp.counter_example_for_attr_implication,
            'test', set(['c']), 0
        )
        self.assertRaises(
            NotCounterExamplePremise,
            self.exp.counter_example_for_attr_implication,
            'test', set(['c', 'b']), 0
        )
        self.assertRaises(
            NotCounterExampleConclusion,
            self.exp.counter_example_for_attr_implication,
            'test', set(['c', 'd', 'b']), 0
        )
        self.assertRaises(
            NotCounterExamplePremise,
            self.exp.counter_example_for_obj_implication,
            'test', set(['1']), 0
        )
        self.assertRaises(
            NotCounterExamplePremise,
            self.exp.counter_example_for_obj_implication,
            'test', set(['1', '2']), 0
        )
        self.assertRaises(
            NotCounterExampleConclusion,
            self.exp.counter_example_for_obj_implication,
            'test', set(['3', '4']), 0
        )
        
    def test_correct_counter_example_for_attribute_implication(self):
        number_of_objects_before = len(self.exp.context.objects)
        self.exp.counter_example_for_attr_implication(
                                    'test', set(['c', 'd']), 0)
        number_of_objects_after = len(self.exp.context.objects)
        self.assertEqual(number_of_objects_after - number_of_objects_before, 
                            1, "Number of objects isn't increased")
                            
    def test_correct_counter_example_for_object_implication(self):
        number_of_attrs_before = len(self.exp.context.attributes)
        self.exp.counter_example_for_obj_implication(
                                    'test', set(['3']), 0)
        number_of_attrs_after = len(self.exp.context.attributes)
        self.assertEqual(number_of_attrs_after - number_of_attrs_before, 
                            1, "Number of attributes isn't increased")

    
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Holds function that read context from tab separated txt file"""

import csv
import fcatng

def read_txt(path):
    """Read context from path, which is tab separated txt file

    Format
    ======

    First line is tab separated attributes' names
    Next an empty line
    Then tab separated 1 and 0, each line corresponds to one object.

    Examples
    ========

    Load example file from tests directory

    >>> c = read_txt('tests/context.txt')
    >>> len(c)
    4
    >>> len(c[0])
    4
    >>> for o in c:
    ...     print o
    ...
    [True, False, False, True]
    [True, False, True, False]
    [False, True, True, False]
    [False, True, True, True]
    >>> print c.objects
    ['g1', 'g2', 'g3', 'g4']
    >>> print c.attributes
    ['a', 'b', 'c', 'd']

    """
    input_file = open(path, "rb")
    rdr = csv.reader(input_file, delimiter="\t")
    rec = next(rdr) # read attributes names

    attributes = []
    for attr in rec:
        attributes.append(str(attr).strip())
    
    next(rdr) # empty line
    
    table = []
    for rec in rdr:
        line = []
        for num in rec:
            if num == "0":
                line.append(False)
            elif num == "1":
                line.append(True)
        table.append(line)
    input_file.close()
    # i + 1 ? 
    objects = ["".join(["g", str(i + 1)]) for i in range(len(table))]

    # TODO: It's hack
    # Strange things happen with csv in case of non standard characters
    if len(attributes) != len(table[0]):
        input_file = open(path, "rb")
        attributes = input_file.readline().split("\t")[:-1]
        input_file.close()

    return fcatng.Context(table, objects, attributes)


def read_mv_txt(path):
    """Read many-valued context from path, which is tab separated txt file

    Format
    ======

    First line is tab separated attributes' names
    Next an empty line
    Then tab separated values, each line corresponds to one object.

    Examples
    ========

    Load example file from tests directory

    >>> c = read_mv_txt('tests/table.txt')
    >>> len(c)
    3
    >>> len(c[0])
    3
    >>> for o in c:
    ...     print o
    ...
    ['7', '6', '7']
    ['7', '2', '9']
    ['1', '3', '4']
    >>> print c.objects
    ['obj1', 'obj2', 'obj3']
    >>> print c.attributes
    ['attr1', 'attr2', 'attr3']

    """
    input_file = open(path, "rb")
    rdr = csv.reader(input_file, delimiter="\t")
    rec = next(rdr) # read objects names
    
    objects = []
    for obj in rec:
        objects.append(str(obj).strip())
    
    rec = next(rdr) # read attributes names

    attributes = []
    for attr in rec:
        attributes.append(str(attr).strip())
    
    next(rdr) # empty line
    
    table = []
    for rec in rdr:
        line = []
        for num in rec:
            line.append(num)
        table.append(line)
    input_file.close()

    # TODO: It's hack
    # Strange things happen with csv in case some of "non standard" characters
    if ((len(table) != 0) and len(attributes) != len(table[0]))\
        or len(table) != len(objects):
        input_file = open(path, "rb")
        objects = input_file.readline().strip().split("\t")
        attributes = input_file.readline().strip().split("\t")
        input_file.close()

    return fcatng.ManyValuedContext(table, objects, attributes)

def uread_mv_txt(path):
    """Read many-valued context from path, which is tab separated txt file

    Format
    ======

    First line is tab separated attributes' names
    Next an empty line
    Then tab separated values, each line corresponds to one object.

    Examples
    ========

    Load example file from tests directory

    >>> c = read_mv_txt('tests/table.txt')
    >>> len(c)
    3
    >>> len(c[0])
    3
    >>> for o in c:
    ...     print o
    ...
    ['7', '6', '7']
    ['7', '2', '9']
    ['1', '3', '4']
    >>> print c.objects
    ['obj1', 'obj2', 'obj3']
    >>> print c.attributes
    ['attr1', 'attr2', 'attr3']

    """
    input_file = open(path, "rb")
    rdr = csv.reader(input_file, delimiter="\t")
    rec = next(rdr) # read objects names
    
    objects = []
    for obj in rec:
        objects.append(str(obj, "utf-8").strip())
    
    rec = next(rdr) # read attributes names

    attributes = []
    for attr in rec:
        attributes.append(str(attr, "utf-8").strip())
    
    next(rdr) # empty line
    
    table = []
    for rec in rdr:
        line = []
        for num in rec:
            line.append(num)
        table.append(line)
    input_file.close()

    # TODO: It's hack
    # Strange things happen with csv in case some of "non standard" characters
    if ((len(table) != 0) and len(attributes) != len(table[0]))\
        or len(table) != len(objects):
        input_file = open(path, "rb")
        objects = input_file.readline().strip().split("\t")
        attributes = input_file.readline().strip().split("\t")
        input_file.close()

    return fcatng.ManyValuedContext(table, objects, attributes)

def write_mv_txt(context, path):
    output_file = open(path, "w")

    output_file.write("\t".join(context.objects))
    output_file.write("\n")

    output_file.write("\t".join(context.attributes))
    output_file.write("\n\n")

    for i in range(len(context.objects)):
        output_file.write("\t".join([str(spam) for spam in context[i]]))
        output_file.write("\n")

    output_file.close()

def uwrite_mv_txt(context, path):
    output_file = open(path, "w")

    output_file.write("\t".join(context.objects).encode("utf-8"))
    output_file.write("\n")

    output_file.write("\t".join(context.attributes).encode("utf-8"))
    output_file.write("\n\n")

    for i in range(len(context.objects)):
        output_file.write("\t".join([str(spam) for spam in context[i]]))
        output_file.write("\n")

    output_file.close()


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    
    




#!/usr/bin/env python
# encoding: utf-8
import copy
import itertools

import fcatng

from fcatng.algorithms.exploration.exploration import (AttributeExploration,
                                                    ExplorationDB)
from fcatng import Implication
# from fcatng.algorithms.closure_operators import simple_closure as closure


class BiUnar(object):

    def  __init__(self, f, g):
        self.f = f
        self.g = g
        
    def __str__(self):
        return '%s%s' % (self.f, self.g)
        
    def automorphic_copy(self):
        return BiUnar(self.g, self.f)
        
        
class Term(object):
    
    def __init__(self, name, function):
        self._name = name
        self._function = function
        
    def __call__(self, bu):
        return self._function(bu)
        
    def __str__(self):
        return self._name
        
        
class Equality(object):
    
    def __init__(self, left_term, right_term):
        self.left = left_term
        self.right = right_term
        
    def __call__(self, bu):
        return self.left(bu) == self.right(bu)
        
    def __str__(self):
        return "%s = %s" % (self.left, self.right)
        
    def automorphic_copy(self):
        return Equality(get_symmetric_term(self.left),
                        get_symmetric_term(self.right))
#        
#    def __eq__(self, eq):
#        return ((self.left == eq.left and self.right == eq.right) or
#                (self.left == eq.right and self.right == eq.left))
#                
#    def __ne__(self, eq):
#        return not self == eq
#
#    def __hash__(self):
#        if self.left < self.right:
#            return hash((self.left, self.right))
#        else:
#            return hash((self.right, self.left))


class CommandLineExpert(object):

    def is_valid(self, imp):
        print "{0}".format(imp)
        return input('Is the following implication valid? Enter "True" or "False": {0}'.format(imp))
                                                                                    
    def explore(self, exploration):
        while exploration.get_open_implications():
            imp = exploration.get_open_implications()[0]
            if self.is_valid(imp):
                exploration.confirm_implications([imp,
                                                get_symmetric_implication(imp)])
            else:
                exploration.reject_implication(imp)

    def provide_counterexample(self, imp):
        print 'Provide a counterexample by typing in two tuples.'
        bu = BiUnar(input('f: '), input('g: '))
        if input('Add as a partial example? Enter "True" or "False": '):
            intent = generate_partial_counterexample(imp, bu)
        else:
            intent = [bu_intent(bu)] * 2    # since our context is partial
        return ((bu, bu.automorphic_copy()),
                (intent, [get_symmetric_attribute_set(s) for s in intent]))


def compose(f, g):
    return tuple([f[x - 1] for x in g])
    
            
def generate_examples(n):
    maps = itertools.product(range(1, n + 1), repeat=n)
    return (BiUnar(mm[0], mm[1]) for mm in itertools.product(maps, repeat=2))
    
    
def bu_intent(bu):
    return set([a for a in attributes if a(bu)])
    
    
def generate_partial_counterexample(imp, bu):
    relevant_attributes = imp.premise | imp.conclusion
    xintent = set([a for a in relevant_attributes if a(bu)])
    qintent = (set(attributes) - relevant_attributes) | xintent
    return xintent, qintent


def find_counter_example(n, imp):
    for e in generate_examples(n):
        if not imp.is_respected(bu_intent(e)):
            return e
    
    
def generate_context(n, attributes):
    objects = [e for e in generate_examples(n)]
    table = [[a(o) for a in attributes] for o in objects]
    cxt = fcatng.partial_context.PartialContext(table,
                                             copy.deepcopy(table),
                                             objects,
                                             attributes)
    # TODO: reduce cxt
    return cxt
    
    
def generate_background_implications(attributes):
    return [Implication(set([x, y]), set([z]))
                    for x in attributes
                    for y in attributes
                    for z in attributes
                    if (x.left == z.left and
                        x.right == y.left and
                        y.right == z.right) or
                       (x.left == z.left and
                        x.right == y.right and
                        y.left == z.right) or
                       (x.left == y.left and
                        x.right == z.left and
                        y.right == z.right)
            ]
            
            
def is_orbit_maximal(eqs):
    symmetric_eqs = get_symmetric_attribute_set(eqs)
    for a in attributes:
        if a in eqs:
            if a not in symmetric_eqs:
                return True
        elif a in symmetric_eqs:
            return False
    return True
    
    
def get_symmetric_attribute_set(eqs):
    symmetric_eqs = set([])
    for e in eqs:
        left = get_symmetric_term(e.left)
        right = get_symmetric_term(e.right)
        if terms.index(left) > terms.index(right):
            left, right = right, left
        for a in attributes:
            if a.left == left and a.right == right:
                symmetric_eqs.add(a)
                break
    return symmetric_eqs
                

def get_symmetric_term(term):
    s = str(term)
    s = s.replace('f', 'F')
    s = s.replace('g', 'f')
    s = s.replace('F', 'g')
    for t in terms:
        if str(t) == s:
            return t
            
            
def get_symmetric_equality(e):
    s = str(e.automorphic_copy())
    for a in attributes:
        if str(a) == s:
            return a


def get_symmetric_implication(imp):
    return Implication(get_symmetric_attribute_set(imp.premise),
                       get_symmetric_attribute_set(imp.conclusion))


terms = [
            Term('id', lambda bu: tuple(range(1, len(bu.f) + 1))),
            Term('f', lambda bu: bu.f),
            Term('g', lambda bu: bu.g),
            Term('ff', lambda bu: compose(bu.f, bu.f)),
            Term('fg', lambda bu: compose(bu.f, bu.g)),
            Term('gf', lambda bu: compose(bu.g, bu.f)),
            Term('gg', lambda bu: compose(bu.g, bu.g))
        ]
        
            
attributes = [Equality(terms[i], terms[j]) for i in range(len(terms))
                                           for j in range(i + 1, len(terms))]
                                              

db = ExplorationDB(generate_context(3, attributes),
                   generate_background_implications(attributes),
                   is_orbit_maximal)
expert = CommandLineExpert()
exploration = AttributeExploration(db, expert)

# expert.explore(exploration)





# -*- coding: utf-8 -*-
"""Reading many-valued contexts from comma-separated text files"""

import csv
import fcatng

def read_mv_csv(path):
    """Read many-valued context from path, which is comma-separated text file

    Format
    ======

    First line consists of comma-separated attribute names.
    Then comma-separated values, each line corresponds to one object.
    The first value in each line is an object name.

    Examples
    ========

    Load example file from tests directory

    >>> c = read_mv_csv('tests/table.csv')
    >>> len(c)
    3
    >>> len(c[0])
    3
    >>> for o in c:
    ...     print o
    ...
    ['7', '6', '7']
    ['7', '2', '9']
    ['1', '3', '4']
    >>> print c.objects
    ['obj1', 'obj2', 'obj3']
    >>> print c.attributes
    ['attr1', 'attr2', 'attr3']

    """
    input_file = open(path, "rb")
    rdr = csv.reader(input_file, delimiter=",")
    
    objects = []
    
    rec = next(rdr) # read attributes names

    attributes = []
    for attr in rec:
        attributes.append(str(attr).strip())
        
    table = []
    for rec in rdr:
        objects.append(str(rec[0]).strip())
        line = []
        for val in rec[1:]:
            line.append(val)
        table.append(line)
    input_file.close()

    return fcatng.ManyValuedContext(table, objects, attributes)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    
    




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest

import fcatng
import fcatng.algorithms.exploration2.exploration as expl
from fcatng import Implication


class ExplorationTest(unittest.TestCase):
    def setUp(self):
        table = [[True, False, False, True],\
            [True, False, True, False],\
            [False, True, True, False],\
            [False, True, True, True]]
        objs = ['1', '2', '3', '4']
        attrs = ['a', 'b', 'c', 'd']
        cxt = expl.ExplorationContext(fcatng.Context(table, objs, attrs))
        self.cxt = cxt

    def test_exploration_object(self):
        e = expl.Exploration(self.cxt)
        session = e.create_session()

equilateral = 'equilateral'
isosceles = 'isosceles'
acute = 'acute-angled'
right = 'right-angled'

class TrianglesTest(unittest.TestCase):

    def setUp(self):
        table = [[False, True, False, False],
                [False, False, True, False],
                [False, False, False, True]]
        objs = [isosceles, acute, right]
        attrs = [equilateral, isosceles, acute, right]
        cxt = expl.ExplorationContext(fcatng.Context(table, objs, attrs))
        self.e = expl.Exploration(cxt)

    def test_all(self):
        basis = [
            Implication(set([acute, right]), set([equilateral, isosceles])),
            Implication(set([equilateral]), set([isosceles, acute]))
        ]
        session = self.e.create_session()
        self.assertEqual(len(session.context.objects), 3)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Derivation and closure operators"""

import copy

def oprime(objects, context):
    """
    Compute the set of all attributes shared by objects in context.
    NB: objects must be of type set
    
    """
    attributes = set(context.attributes[:])
    for o in objects:
        attributes &= context.get_object_intent(o)
    return attributes
    

def aprime(attributes, context):
    """
    Compute the set of all objects sharing attributes in context.
    NB: attributes must be of type set
    
    """
    objects = set(context.objects[:])
    for a in attributes:
        objects &= context.get_attribute_extent(a)
    return objects
    
    
def oclosure(objects, context):
    """Return the closure of objects in context as a sorted list"""
    return sorted(aprime(oprime(objects, context), context))
    

def aclosure(attributes, context):
    """Return the closure of attributes in context as a sorted list"""
    return sorted(oprime(aprime(attributes, context), context))
    
    
def simple_closure(s, implications):
    """
    Input:  A set of implications and an attribute set s
    Output: The closure of s with respect to implications
    
    Examples
    ========
    
    >>> from fcatng.implication import Implication
    >>> cd2a = Implication(set(('c', 'd')), set(('a')))
    >>> ad2c = Implication(set(('a', 'd')), set(('c')))
    >>> ab2cd = Implication(set(('a', 'b')), set(('c', 'd')))
    >>> imps = [cd2a, ad2c, ab2cd]
    >>> print simple_closure(set('a'), imps)
    set(['a'])
    >>> print simple_closure(set(), imps)
    set([])
    >>> simple_closure(set(['b', 'c', 'd']), imps) == set(['a', 'b', 'c', 'd'])
    True
    
    >>> a2bc = Implication(set(('a')), set(('b', 'c')))
    >>> ce2abd = Implication(set(('c', 'e')), set(('a', 'b', 'd')))
    >>> de2abc = Implication(set(('d', 'e')), set(('a', 'b', 'c')))
    >>> cd2abe = Implication(set(('c', 'd')), set(('a', 'b', 'e')))
    >>> imps = [a2bc, ce2abd, de2abc, cd2abe]
    >>> simple_closure(set(['b', 'a']), imps) == set(['a', 'b', 'c'])
    True
    >>> simple_closure(set(['a', 'e']), imps) == set(['a', 'b', 'c', 'd', 'e'])
    True
    >>> imps = [ce2abd, a2bc, de2abc, cd2abe]
    >>> simple_closure(set(['a', 'e']), imps) == set(['a', 'b', 'c', 'd', 'e'])
    True
    
    """
    unused_imps = implications[:]
    new_closure = s.copy()
    changed = True
    while changed:
        changed = False
        for imp in unused_imps:
            if imp.premise <= new_closure:
                new_closure |= imp.conclusion
                changed = True
                unused_imps.remove(imp)
    return new_closure
    

def lin_closure(s, implications):
    """
    Input:  A set of implications and an attribute set s
    Output: The closure of s with respect to implications
    NB: This implementation is not linear-time.
        Use lists instead of dicts to get a liner-time implementation.
    
    Examples
    ========
    
    >>> from fcatng.implication import Implication
    >>> cd2a = Implication(set(('c', 'd')), set(('a')))
    >>> ad2c = Implication(set(('a', 'd')), set(('c')))
    >>> ab2cd = Implication(set(('a', 'b')), set(('c', 'd')))
    >>> imps = [cd2a, ad2c, ab2cd]
    >>> print lin_closure(set('a'), imps)
    set(['a'])
    >>> print lin_closure(set(), imps)
    set([])
    >>> lin_closure(set(['b', 'c', 'd']), imps) == set(['a', 'b', 'c', 'd'])
    True
    
    >>> a2bc = Implication(set(('a')), set(('b', 'c')))
    >>> ce2abd = Implication(set(('c', 'e')), set(('a', 'b', 'd')))
    >>> de2abc = Implication(set(('d', 'e')), set(('a', 'b', 'c')))
    >>> cd2abe = Implication(set(('c', 'd')), set(('a', 'b', 'e')))
    >>> imps = [a2bc, ce2abd, de2abc, cd2abe]
    >>> lin_closure(set(['b', 'a']), imps) == set(['a', 'b', 'c'])
    True
    >>> lin_closure(set(['a', 'e']), imps) == set(['a', 'b', 'c', 'd', 'e'])
    True
    >>> imps = [ce2abd, a2bc, de2abc, cd2abe]
    >>> lin_closure(set(['a', 'e']), imps) == set(['a', 'b', 'c', 'd', 'e'])
    True
    
    """
    new_closure = s.copy()
    count, imps = {}, {}
    for imp in implications:
        count[imp] = len(imp.premise)
        if count[imp] == 0:
            new_closure |= imp.conclusion
        for a in imp.premise:
            imps.setdefault(a, [])
            imps[a].append(imp)
    update = list(new_closure)
    
    while update:
        m = update[-1]
        del update[-1]
        imps.setdefault(m, [])
        for imp in imps[m]:
            count[imp] -= 1
            if count[imp] == 0:
                add = imp.conclusion - new_closure
                new_closure |= add
                update.extend(add)
    return new_closure


def closure(current, base_set, implications, prefLen):
    """
    return the closure of attributes
    NB: current and base_set must be of type set
    """
    unused_imps = copy.copy(implications)
    old_closure = copy.copy(base_set)
    new_closure = copy.copy(current)
    
    while (old_closure != new_closure):
        old_closure = copy.copy(new_closure)
        delete_list = [];
        
        for imp in unused_imps:
            if imp.get_premise() <= new_closure:
                new_closure |= imp.get_conclusion()
                
                for x in base_set[:prefLen]:
                    if (((x in new_closure) and not (x in current)) 
                        or (not (x in new_closure) and (x in current))):
                        return False, set()
                
                delete_list.append(imp)
                
        for imp in delete_list:
            unused_imps.remove(imp)
    
    return True, new_closure    
                

if __name__ == "__main__":
    import doctest
    doctest.testmod()



""" Comparing contexts """


def subseteq_table(table_left, table_right):
    """
    Checks whether one binary relation of two sets (presented as  bool table) is
    subset or equals another. Tables should be given as lists of bool lists and
    should have the same dimensions. Result is bool.
    table_left ?\subseteq table_right
    """
    if len(table_left) != len(table_right):
        raise ValueError("Number of rows in left table (=%i) and number of rows"
                         "in right table (=%i) must agree" % (len(table_left),
                                                              len(table_right)))
    elif (len(table_left) != 0) and len(table_left[0]) != len(table_right[0]):
        raise ValueError("Number of columns in left table (=%i) and number of"
                         "columns in right table (=%i) must agree"
                         %(len(table_left[0]), len(table_right[0])))

    row_length = len(table_left)
    table_size = row_length * len(table_left[0])
    for i in range(table_size):
        a = (i / row_length)
        b = (i % row_length)
        if not (table_left[b][a] <= table_right[b][a]):
            return False

    return True


# -*- coding: utf-8 -*-
"""Holds functions that read a concept system from .xml file"""

from xml.dom.minidom import getDOMImplementation
import xml.parsers.expat

import fcatng

def write_xml(path, cs):
    """Write concept system to xml document
    
    Examples
    ========
    
    >>> from fcatng import ConceptLattice
    >>> c = fcatng.read_cxt('tests/context.cxt')
    >>> cs = ConceptLattice(c)
    >>> write_xml("tests/test.xml", cs)
    """
    objects = set()
    attributes = set()
    for concept in cs:
        objects = objects.union(concept.extent)
        attributes = attributes.union(concept.intent)
    objects = list(objects)
    attributes = list(attributes)
    
    objects_ids = dict([(objects[i], "o{0}".format(i+1))  for i in range(len(objects))])
    attributes_ids = dict([(attributes[i], "a{0}".format(i+1))  for i in range(len(attributes))])
    
    out = file(path, "wb")

    impl = getDOMImplementation()

    newdoc = impl.createDocument(None, "conceptsystem", None)
    top_element = newdoc.documentElement
    
    element = newdoc.createElement("objects")
    for obj in objects:
        obj_element = newdoc.createElement("object")
        obj_element.setAttribute("id", objects_ids[obj])
        textnode = newdoc.createTextNode(obj.encode("utf-8"))
        obj_element.appendChild(textnode)
        element.appendChild(obj_element)
    top_element.appendChild(element)
    
    element = newdoc.createElement("attributes")
    for attr in attributes:
        attr_element = newdoc.createElement("attribute")
        attr_element.setAttribute("id", attributes_ids[attr])
        textnode = newdoc.createTextNode(attr.encode("utf-8"))
        attr_element.appendChild(textnode)
        element.appendChild(attr_element)
    top_element.appendChild(element)
    
    element = newdoc.createElement("concepts")
    for concept in cs:
        c_element = newdoc.createElement("concept")
        
        e_element = newdoc.createElement("extent")
        for obj in concept.extent:
            obj_element = newdoc.createElement("object")
            obj_element.setAttribute("ref", objects_ids[obj])
            e_element.appendChild(obj_element)
            
        c_element.appendChild(e_element)
        
        i_element = newdoc.createElement("intent")
        for attr in concept.intent:
            attr_element = newdoc.createElement("attribute")
            attr_element.setAttribute("ref", attributes_ids[attr])
            i_element.appendChild(attr_element)
            
        c_element.appendChild(i_element)
            
        m_element = newdoc.createElement("meta")
        for key in list(concept.meta.keys()):
            m_element.setAttribute(str(key.replace(" ", "_")), str(concept.meta[key]))
            
        c_element.appendChild(m_element)
        
        element.appendChild(c_element)
        
    top_element.appendChild(element)
    
    newdoc.writexml(out, indent="\n", addindent="\t", encoding="UTF-8")
    out.close()


def read_xml(path):
    """Read concept system from valid xml file.
    
    Examples
    ========
    
    >>> cs = read_xml('tests/concepts.xml')
    >>> print cs
    (['obj1', 'obj2', 'obj3'], ['attr1'])
    (['obj2', 'obj3'], ['attr1', 'attr2'])
    (['obj1'], ['attr1', 'attr3'])
    ([], M)
    """
    global new_obj, new_attr, cs, buffer
    
    buffer = ""
    
    cs = fcatng.ConceptSystem()
    
    new_obj = None
    new_attr = None
    
    objects = []
    d_objects = {}
    
    attributes = []
    d_attributes = {}
    
    new_intent = []
    new_extent = []
    new_meta = {}
    
    def start_element(name, attrs):
        global new_obj, new_attr, buffer
        global new_extent, new_intent, new_meta
        if name == "object":
            if "id" in list(attrs.keys()):
                buffer = ""
                new_obj = attrs["id"]
            elif "ref" in list(attrs.keys()):
                new_extent.append(d_objects[attrs["ref"]])
        elif name == "attribute":
            if "id" in list(attrs.keys()):
                buffer = ""
                new_attr = attrs["id"]
            elif "ref" in list(attrs.keys()):
                new_intent.append(d_attributes[attrs["ref"]])
        elif name == "meta":
            for key in list(attrs.keys()):
                new_meta[str(key).replace("_", " ")] = float(attrs[key])
        elif name == "concept":
            new_intent = []
            new_extent = []
            new_meta = {}
        
    def end_element(name):
        global cs, new_intent, new_extent, new_meta
        global new_obj, new_attr, buffer
        if name == "object":
            if new_obj:
                d_objects[new_obj] = buffer
                objects.append(buffer)
                new_obj = None
                buffer = ""
        elif name == "attribute":
            if new_attr:
                d_attributes[new_attr] = buffer
                attributes.append(buffer)
                new_attr = None
                buffer = ""
        elif name == "concept":
            new_concept = fcatng.Concept(new_extent, new_intent)
            new_concept.meta = new_meta
            cs.append(new_concept)
            
            new_extent = []
            new_intent = []
            new_meta = {}
    
    def char_data(data):
        global buffer
        if data[0] == "\n":
            return
        data = data.strip()
        buffer += data
    
    p = xml.parsers.expat.ParserCreate()
    
    p.StartElementHandler = start_element
    p.EndElementHandler = end_element
    p.CharacterDataHandler = char_data
    
    f = open(path)
    p.ParseFile(f)
    
    return cs
    
def uread_xml(path):
    global new_obj, new_attr, cs, buffer

    buffer = ""

    cs = fcatng.ConceptSystem()

    new_obj = None
    new_attr = None

    objects = []
    d_objects = {}

    attributes = []
    d_attributes = {}

    new_intent = []
    new_extent = []
    new_meta = {}

    def start_element(name, attrs):
        global new_obj, new_attr, buffer
        global new_extent, new_intent, new_meta
        if name == "object":
            if "id" in list(attrs.keys()):
                buffer = ""
                new_obj = attrs["id"]
            elif "ref" in list(attrs.keys()):
                new_extent.append(d_objects[attrs["ref"]])
        elif name == "attribute":
            if "id" in list(attrs.keys()):
                buffer = ""
                new_attr = attrs["id"]
            elif "ref" in list(attrs.keys()):
                new_intent.append(d_attributes[attrs["ref"]])
        elif name == "meta":
            for key in list(attrs.keys()):
                new_meta[str(key).replace("_", " ")] = float(attrs[key])
        elif name == "concept":
            new_intent = []
            new_extent = []
            new_meta = {}

    def end_element(name):
        global cs, new_intent, new_extent, new_meta
        global new_obj, new_attr, buffer
        if name == "object":
            if new_obj:
                d_objects[new_obj] = buffer
                objects.append(buffer)
                new_obj = None
                buffer = ""
        elif name == "attribute":
            if new_attr:
                d_attributes[new_attr] = buffer
                attributes.append(buffer)
                new_attr = None
                buffer = ""
        elif name == "concept":
            new_concept = fcatng.Concept(new_extent, new_intent)
            new_concept.meta = new_meta
            cs.append(new_concept)

            new_extent = []
            new_intent = []
            new_meta = {}

    def char_data(data):
        global buffer
        if data[0] == "\n":
            return
        data = data.strip()
        buffer += data

    p = xml.parsers.expat.ParserCreate(encoding="UTF-8")

    p.StartElementHandler = start_element
    p.EndElementHandler = end_element
    p.CharacterDataHandler = char_data
    
    f = open(path)
    p.ParseFile(f)

    return cs


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""Holds function that write lattice diagram to .dot file""" 

import fcatng

def write_dot(cs, path):
    """For given concept_system writes a .dot file, which contains graph
    representation of lattice diagram in format understandable 
    by graphviz' dot. "path" contains name of the output file.  
 
    Examples
    ========

    Load example file from tests directory

    >>> c = fcatng.read_cxt('tests/context.cxt')
    >>> cs = fcatng.norris(c)
    >>> write_dot(cs, "")
    Traceback (most recent call last):
        ...
    AssertionError: Filename can't be empty

    """
    assert len(path)!=0, "Filename can't be empty"
    output_file = open(path, "w")
    output_file.write("digraph L{")
    output_file.write("\n")
    output_file.write("nodesep=.25;\n")
    output_file.write("node[shape=circle,label=\"\"];")
    output_file.write("\n")
    output_file.write("edge[dir=\"none\",minlen=2,fontsize=11,labelfloat=true,fontname=Arial];")
    output_file.write("\n")

    own_objects = find_own_objects(cs)
    own_attributes = find_own_attributes(cs)
    for i in range(len(cs)):
        output_file.write("c%i [width=0.25]\n" % i)
        if len(own_objects[cs[i]])!=0:
            output_file.write("c%i -> c%i" % (i, i))
            # TODO:
            if len(own_objects[cs[i]]) >= 5:
                own_objects[cs[i]] = [str(len(own_objects[cs[i]]))]
            output_file.write("[headlabel=\"%s\", " %\
                     "\\n ".join(own_objects[cs[i]]))
            output_file.write(
                "labeldistance=1,labelangle=270,color=transparent]\n")
        if len(own_attributes[cs[i]])!=0:
            output_file.write("c%i -> c%i" % (i, i))
            # TODO:
            if len(own_attributes[cs[i]]) >= 5:
                own_attributes[cs[i]] = [str(len(own_attributes[cs[i]]))]
            output_file.write("[taillabel=\"%s\", " %\
                     "\\n ".join(own_attributes[cs[i]]))
            output_file.write(
                "labeldistance=2,labelangle=90,color=transparent]\n")

    parents = fcatng.compute_covering_relation(cs)
    for i in range(len(cs)):
        for p in parents[cs[i]]:
            output_file.write("c%i -> c%i\n" % (cs.index(p), i))
    output_file.write("}")

    output_file.close()

def uwrite_dot(cs, path, full=False):
    assert len(path)!=0, "Filename can't be empty"
    
    import codecs
    
    output_file = codecs.open(path, "w", "utf-8")
    output_file.write("digraph L{")
    output_file.write("\n")
    output_file.write("nodesep=.25;\n")
    output_file.write("node[shape=circle,label=\"\"];")
    output_file.write("\n")
    output_file.write("edge[dir=\"none\",minlen=2,fontsize=9,labelfloat=true,fontname=Arial,color=grey,style=\"setlinewidth(.7)\"];")
    output_file.write("\n")

    own_objects = find_own_objects(cs)
    own_attributes = find_own_attributes(cs)

    for i in range(len(cs)):
        output_file.write("c%i [width=0.25]\n" % i)

    parents = fcatng.compute_covering_relation(cs)
    for i in range(len(cs)):
        for p in parents[cs[i]]:
            output_file.write("c%i -> c%i\n" % (cs.index(p), i))

    for i in range(len(cs)):
        if len(own_objects[cs[i]])!=0:
            output_file.write("c%i -> c%i" % (i, i))
            # TODO:
            if not full:
                if len(own_objects[cs[i]]) >= 1:
                    own_objects[cs[i]] = [str(len(own_objects[cs[i]]))]
            if len(own_objects[cs[i]]) > 6:
                own_objects[cs[i]] = own_objects[cs[i]][:6] + ["..."]
            n = len(own_objects[cs[i]])
            output_file.write("[headlabel=\"%s\", " %\
                     "\\n ".join(own_objects[cs[i]]))
            output_file.write(
                "labeldistance={0},labelangle=270,color=transparent]\n".format(n/1.8 + 0.5))
        if len(own_attributes[cs[i]])!=0:
            output_file.write("c%i -> c%i" % (i, i))
            # TODO:
            if not full:
                if len(own_attributes[cs[i]]) >= 1:
                    own_attributes[cs[i]] = [str(len(own_attributes[cs[i]]))]
            if len(own_attributes[cs[i]]) > 6:
                own_attributes[cs[i]] = own_attributes[cs[i]][:6] + ["..."]
            n = len(own_attributes[cs[i]])
            output_file.write("[taillabel=\"%s\", " %\
                     "\\n ".join(own_attributes[cs[i]]))
            output_file.write(
                "labeldistance={0},labelangle=90,color=transparent]\n".format(n/1.8 + 0.5))

    output_file.write("}")

    output_file.close()

def find_own_objects(cs):
    """Return set of own objects for current concept"""
    own_objects = {}
    for con in cs:
        own_objects[con] = []
        for obj in con.extent:
            own_objects[con].append(obj)
            for sub_con in cs:
                if sub_con.extent < con.extent and\
                        obj in sub_con.extent:
                    own_objects[con].pop()
                    break
    return own_objects

def find_own_attributes(cs):
    """Return set of own attributes for current concept"""
    own_attributes = {}
    for con in cs:
        own_attributes[con] = []
        for attr in con.intent:
            own_attributes[con].append(attr)
            for sub_con in cs:
                if sub_con.intent < con.intent and\
                        attr in sub_con.intent:
                    own_attributes[con].pop()
                    break
    return own_attributes


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8
from copy import copy, deepcopy
from fcatng import Context
from fcatng.partial_context import PartialContext

class ExplorationException(Exception):
    pass
    
class NotCounterexample(ExplorationException):
    pass
    
class IllegalContextModification(ExplorationException):
    
    def __str__(self):
        return "Attempted context modification conflicts with existing background knowledge"
    
class NotUniqueObjectName(ExplorationException):
    pass
    
class NotUniqueAttributeName(ExplorationException):
    pass
    
def context_modifier(F):
    """
    Decorator for methods where context is somehow modified.
    Recompute implications in context.
    """
    def wrapper(db, *args):
        F(db, *args)
        db._cxt_implications = db._cxt.get_attribute_implications(
                                            confirmed=db._implications,
                                            cond=db._cond)
    return wrapper
    
def base_modifier(F):
    """
    Decorator for methods where background knowledge is somehow modified.
    Recomputes context relative basis
    """
    def wrapper(db, *args):
        F(db, *args)
        db._cxt_implications = db._cxt.get_attribute_implications(
                                                confirmed=db._implications,
                                                cond=db._cond)
    return wrapper
    
class ExplorationDB(object):
    """docstring for ExplorationDB"""
    def __init__(self, context, implications, cond=lambda x: True):
        super(ExplorationDB, self).__init__()
        self._cxt = context
        # background knowledge
        self._implications = implications
        # condition on pseudo-intents
        self._cond = cond
        # relative basis
        self._cxt_implications = context.get_attribute_implications(
                                                confirmed=self._implications,
                                                cond=cond)
    
    def _check_intents(self, intents):
        for intent in intents:
            for imp in self._implications:
                if not imp.is_respected(intent):
                    raise IllegalContextModification()

    @base_modifier
    def confirm_implication(self, imp):
        """docstring for confirm_implication"""
        self._implications.append(imp)
        
    @base_modifier
    def unconfirm_implication(self, imp):
        self._implications.remove(imp)
    
    @context_modifier
    def add_example(self, name, intent):
        self._check_intents([intent])

        if name in self._cxt.objects:
            raise NotUniqueObjectName()
        self._cxt.add_object_with_intent(intent, name)
        
    @context_modifier
    def delete_example(self, name):
        self._cxt.delete_object_by_name(name)
    
    @context_modifier
    def edit_example(self, name, old_name, intent):
        old_intent = self._cxt.get_object_intent(old_name)
        if intent != old_intent:
            self._check_intents([intent])
            self._cxt.set_object_intent(intent, old_name)
        if old_name != name:
            if name in self._cxt.objects:
                raise NotUniqueObjectName()
            else:
                self._cxt.rename_object(old_name, name)
        
    @context_modifier
    def add_attribute(self, name, extent):
        if name in self._cxt.attributes:
            raise NotUniqueAttributeName()
        self._cxt.add_attribute_with_extent(extent, name)
        
    @context_modifier
    def delete_attribute(self, name):
        self._cxt.delete_attribute_by_name(name)

    @context_modifier
    def touch(self):
        pass
        
    @context_modifier
    def edit_attribute(self, name, old_name, extent):
        intents = []
        for obj in self._cxt.objects:
            if obj in extent:
                intents.append(self._cxt.get_object_intent(obj).add(old_name))
            else:
                intents.append(self._cxt.get_object_intent(obj).discard(old_name))
        self._check_intents(intents)

        first = False
        for attribute in self._cxt.attributes:
            if name == attribute:
                if first:
                    raise NotUniqueAttributeName()
                else:
                    first = True
        self._cxt.set_attribute_extent(extent, old_name)
        self._cxt.rename_attribute(old_name, name)
            
    def get_open_implications(self):
        return copy(self._cxt_implications)
        
    def get_base(self):
        return copy(self._implications)
        
    def get_object_names(self):
        return self._cxt.objects
        
    def get_attribute_names(self):
        return self._cxt.attributes
        
    def complete(self):
        # TODO: refactor
        if type(self._cxt) == PartialContext:
            self._cxt.complete(self._implications)
    
    objects = property(get_object_names)
    attributes = property(get_attribute_names)
    open_implications = property(get_open_implications)
    base = property(get_base)
        

class AttributeExploration(object):
    """Exploration"""
    def __init__(self, db, expert):
        super(AttributeExploration, self).__init__()
        self.db = db
        self.expert = expert
        
    def confirm_implication(self, imp):
        """docstring for confirm_implication"""
        self.confirm_implications([imp])
        
    def confirm_implications(self, imps):
        """docstring for confirm_implication"""
        for i in imps:
            self.db.confirm_implication(i)
        self.db.complete()

    def reject_implication(self, imp):
        """docstring for reject_implication"""
        examples, intents = self.expert.provide_counterexample(imp)
        # TODO: refactor
        if type(intents) == set:
            examples, intents = [examples], [intents]
        if imp.is_respected(intents[0]):
            raise NotCounterexample()
        for i in range(len(examples)):
            self.db.add_example(examples[i], intents[i])
        self.db.complete()
            
    def unconfirm_implication(self, imp):
        self.db.unconfirm_implication(imp)
            
    def get_open_implications(self):
        return self.db.open_implications



# -*- coding: utf-8 -*-
"""Module holds functions that read context from various formats of files"""

from fcatng.readwrite.tab_separated import *
from fcatng.readwrite.comma_separated import *
from fcatng.readwrite.cxt import *
from fcatng.readwrite.dot import *
from fcatng.readwrite.xml_ import *
from fcatng.readwrite.fimi import *



# -*- coding: utf-8 -*-
"""Holds a function that computes covering relation for a given concept
system"""

def compute_covering_relation(cs):
    """Computes covering relation for a given concept system.
    
    Returns a dictionary containing sets of parents for each concept.

    Examples
    ========

    >>> from fcatng import *
    >>> ct = [[True, False, False, True],\
              [True, False, True, False],\
              [False, True, True, False],\
              [False, True, True, True]]
    >>> objs = [1, 2, 3, 4]
    >>> attrs = ['a', 'b', 'c', 'd']
    >>> c = Context(ct, objs, attrs)
    >>> cs = norris(c)
    >>> print cs
    ([], M)
    ([1], ['a', 'd'])
    ([2], ['a', 'c'])
    ([1, 2], ['a'])
    ([3, 4], ['b', 'c'])
    ([2, 3, 4], ['c'])
    (G, [])
    ([4], ['b', 'c', 'd'])
    ([1, 4], ['d'])
    >>> parents = compute_covering_relation(cs)
    >>> for c in cs:
    ...     print c
    ...     for p in parents[c]:
    ...         print ' '.join(['<<<', str(p)])
    ...
    ([], M)
    <<< ([1], ['a', 'd'])
    <<< ([2], ['a', 'c'])
    <<< ([4], ['b', 'c', 'd'])
    ([1], ['a', 'd'])
    <<< ([1, 2], ['a'])
    <<< ([1, 4], ['d'])
    ([2], ['a', 'c'])
    <<< ([1, 2], ['a'])
    <<< ([2, 3, 4], ['c'])
    ([1, 2], ['a'])
    <<< (G, [])
    ([3, 4], ['b', 'c'])
    <<< ([2, 3, 4], ['c'])
    ([2, 3, 4], ['c'])
    <<< (G, [])
    (G, [])
    ([4], ['b', 'c', 'd'])
    <<< ([3, 4], ['b', 'c'])
    <<< ([1, 4], ['d'])
    ([1, 4], ['d'])
    <<< (G, [])

    """
    parents = dict([(c, set()) for c in cs])

    for i in range(len(cs)):
        for j in range(len(cs)):
            if cs[i].intent < cs[j].intent:
                parents[cs[j]].add(cs[i])
                for k in range(len(cs)):
                    if cs[i].intent < cs[k].intent and\
                       cs[k].intent < cs[j].intent:
                            parents[cs[j]].remove(cs[i])
                            break
    return parents


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""FCA algorithms"""

from fcatng.algorithms.norris import *
from fcatng.algorithms.covering_relation import *
from fcatng.algorithms.scaling import *
from fcatng.algorithms.filters import *
from fcatng.algorithms.dg_basis import compute_dg_basis

# from .filters import (filter_concepts, compute_index)
# from .probability import compute_probability
# from .separation import compute_separation_index
# from .stability import (compute_estability, compute_istability)
# from .extentsize import compute_extent_size
# from .intentsize import compute_intent_size

# def get_compute_functions():
#     functions = {"Probability index" : compute_probability,
#                  "Separation index" : compute_separation_index,
#                  "Intensional stability" : compute_istability,
#                  "Extensional stability" : compute_estability,
#                  "Extent size" : compute_extent_size,
#                  "Intent size" : compute_intent_size}
#     return functions

# def get_modes():
#     return ["part", "abs", "value"]



# -*- coding: utf-8 -*-
"""Holds ConceptSystem class"""

from .concept import Concept
#from algorithms import compute_covering_relation

class ConceptSystem(object):
    """A ConceptSystem class contains a set of concepts

    Emulates container

    Examples
    ========
    
    >>> c = Concept([1, 2], ['a', 'b'])
    >>> cs = ConceptSystem([c])
    >>> c in cs
    True
    >>> Concept([1], ['c']) in cs
    False
    >>> print cs
    ([1, 2], ['a', 'b'])

    """
    def get_top_concept(self):
        # TODO: change
        return [c for c in self._concepts if not self.filter(c)][0]

    top_concept = property(get_top_concept)

    def get_bottom_concept(self):
        # TODO: change
        return [c for c in self._concepts if not self.ideal(c)][0]

    bottom_concept = property(get_bottom_concept)

    def filter(self, concept):
        # TODO: optimize
        return [c for c in self._concepts if concept.intent > c.intent]

    def ideal(self, concept):
        # TODO: optimize
        return [c for c in self._concepts if c.intent > concept.intent]

    def __init__(self, concepts=[]):
        self._concepts = concepts[:]
        self._parents = None
        self._bottom_concept = None
        self._top_concept = None

    def __len__(self):
        return len(self._concepts)

    def __getitem__(self, key):
        return self._concepts[key]

    def __contains__(self, value):
        return value in self._concepts

    def __str__(self):
        s = ""
        for c in self._concepts:
            s = s + "%s\n" % str(c)
        return s[:-1]

    def index(self, concept):
        return self._concepts.index(concept)

    def append(self, concept):
        if isinstance(concept, Concept):
            self._concepts.append(concept)
        else:
            raise TypeError("concept must be an instance of the Concept class")
        self._parents = None
        # TODO: optimize

    def remove(self, concept):
        if isinstance(concept, Concept):
            self._concepts.remove(concept)
        self._parents = None

    def compute_covering_relation(self):
        """Computes covering relation for a given concept system.

        Returns a dictionary containing sets of parents for each concept.

        Examples
        ========

        """
        cs = self
        parents = dict([(c, set()) for c in cs])

        for i in range(len(cs)):
            for j in range(len(cs)):
                if cs[i].intent < cs[j].intent:
                    parents[cs[j]].add(cs[i])
                    for k in range(len(cs)):
                        if cs[i].intent < cs[k].intent and\
                           cs[k].intent < cs[j].intent:
                                parents[cs[j]].remove(cs[i])
                                break
        return parents

    def parents(self, concept):
        if not self._parents:
            self._parents = self.compute_covering_relation()
        return self._parents[concept]

    def children(self, concept):
        return set([c for c in self._concepts if concept in self.parents(c)])


if __name__ == "__main__":
    import doctest
    doctest.testmod()
